import asyncio
from threading import RLock
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union

import aiohttp
from pysmartthings import (
    Attribute,
    Capability,
    Command,
    DeviceEntity,
    DeviceStatus,
    Location,
    Room,
    SmartThings,
)

from platypush.entities import (
    DimmerEntityManager,
    Entity,
    EnumSwitchEntityManager,
    LightEntityManager,
    SensorEntityManager,
    SwitchEntityManager,
)
from platypush.entities.devices import Device
from platypush.entities.dimmers import Dimmer
from platypush.entities.lights import Light
from platypush.entities.sensors import Sensor
from platypush.entities.switches import EnumSwitch, Switch
from platypush.plugins import RunnablePlugin, action
from platypush.utils import camel_case_to_snake_case

from ._mappers import DeviceMapper, device_mappers


# pylint: disable=too-many-ancestors
class SmartthingsPlugin(
    RunnablePlugin,
    DimmerEntityManager,
    EnumSwitchEntityManager,
    LightEntityManager,
    SensorEntityManager,
    SwitchEntityManager,
):
    """
    Plugin to interact with devices and locations registered to a Samsung SmartThings account.
    """

    _timeout = aiohttp.ClientTimeout(total=20.0)

    def __init__(
        self, access_token: str, poll_interval: Optional[float] = 20.0, **kwargs
    ):
        """
        :param access_token: SmartThings API access token - you can get one at https://account.smartthings.com/tokens.
        :param poll_interval: How often the plugin should poll for changes, in seconds (default: 20).
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self._access_token = access_token
        self._refresh_lock = RLock()
        self._execute_lock = RLock()

        self._locations: List[Location] = []
        self._devices: List[DeviceEntity] = []
        self._rooms_by_location: Dict[str, Room] = {}

        self._locations_by_id: Dict[str, Location] = {}
        self._locations_by_name: Dict[str, Location] = {}
        self._devices_by_id: Dict[str, DeviceEntity] = {}
        self._devices_by_name: Dict[str, DeviceEntity] = {}
        self._rooms_by_id: Dict[str, Room] = {}
        self._rooms_by_location_and_id: Dict[str, Dict[str, Room]] = {}
        self._rooms_by_location_and_name: Dict[str, Dict[str, Room]] = {}
        self._entities_by_id: Dict[str, Entity] = {}

    async def _refresh_locations(self, api):
        self._locations = await api.locations()
        self._locations_by_id = {loc.location_id: loc for loc in self._locations}
        self._locations_by_name = {loc.name: loc for loc in self._locations}

    async def _refresh_devices(self, api):
        self._devices = await api.devices()
        self._devices_by_id = {dev.device_id: dev for dev in self._devices}
        self._devices_by_name = {dev.label: dev for dev in self._devices}

    async def _refresh_rooms(self, api, location_id: str):
        self._rooms_by_location[location_id] = await api.rooms(location_id=location_id)

        self._rooms_by_id.update(
            **{room.room_id: room for room in self._rooms_by_location[location_id]}
        )

        self._rooms_by_location_and_id[location_id] = {
            room.room_id: room for room in self._rooms_by_location[location_id]
        }

        self._rooms_by_location_and_name[location_id] = {
            room.name: room for room in self._rooms_by_location[location_id]
        }

    async def _refresh_info(self):
        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            api = SmartThings(session, self._access_token)
            tasks = [
                asyncio.ensure_future(self._refresh_locations(api)),
                asyncio.ensure_future(self._refresh_devices(api)),
            ]

            await asyncio.gather(*tasks)

            room_tasks = [
                asyncio.ensure_future(self._refresh_rooms(api, location.location_id))
                for location in self._locations
            ]

            await asyncio.gather(*room_tasks)

    def refresh_info(self):
        with self._refresh_lock:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                loop.run_until_complete(self._refresh_info())
            finally:
                loop.stop()

    def _location_to_dict(self, location) -> Dict:
        return {
            'name': location.name,
            'location_id': location.location_id,
            'country_code': location.country_code,
            'locale': location.locale,
            'latitude': location.latitude,
            'longitude': location.longitude,
            'temperature_scale': location.temperature_scale,
            'region_radius': location.region_radius,
            'timezone_id': location.timezone_id,
            'rooms': {
                room.room_id: self._room_to_dict(room)
                for room in self._rooms_by_location.get(location.location_id, {})
            },
        }

    @staticmethod
    def _device_to_dict(device) -> Dict:
        return {
            'capabilities': device.capabilities,
            'name': device.label,
            'device_id': device.device_id,
            'location_id': device.location_id,
            'room_id': device.room_id,
            'device_type_id': device.device_type_id,
            'device_type_name': device.device_type_name,
            'device_type_network': device.device_type_network,
        }

    @staticmethod
    def _room_to_dict(room) -> Dict:
        return {
            'name': room.name,
            'background_image': room.background_image,
            'room_id': room.room_id,
            'location_id': room.location_id,
        }

    @action
    def info(self) -> Dict[str, Dict[str, dict]]:
        """
        Return the objects registered to the account, including locations and devices.

          .. code-block:: json

            {
                "devices": {
                    "smart-tv-id": {
                        "capabilities": [
                            "ocf",
                            "switch",
                            "audioVolume",
                            "audioMute",
                            "tvChannel",
                            "mediaInputSource",
                            "mediaPlayback",
                            "mediaTrackControl",
                            "custom.error",
                            "custom.picturemode",
                            "custom.soundmode",
                            "custom.accessibility",
                            "custom.launchapp",
                            "custom.recording",
                            "custom.tvsearch",
                            "custom.disabledCapabilities",
                            "samsungvd.ambient",
                            "samsungvd.ambientContent",
                            "samsungvd.ambient18",
                            "samsungvd.mediaInputSource",
                            "refresh",
                            "execute",
                            "samsungvd.firmwareVersion",
                            "samsungvd.supportsPowerOnByOcf"
                        ],
                        "device_id": "smart-tv-id",
                        "device_type_id": null,
                        "device_type_name": null,
                        "device_type_network": null,
                        "location_id": "location-id",
                        "name": "Samsung Smart TV",
                        "room_id": "room-1"
                    },
                    "tv-switch-id": {
                        "capabilities": [
                            "switch",
                            "refresh",
                            "healthCheck"
                        ],
                        "device_id": "tv-switch-id",
                        "device_type_id": null,
                        "device_type_name": null,
                        "device_type_network": null,
                        "location_id": "location-id",
                        "name": "TV Smart Switch",
                        "room_id": "room-1"
                    },
                    "lights-switch-id": {
                        "capabilities": [
                            "switch",
                            "refresh",
                            "healthCheck"
                        ],
                        "device_id": "lights-switch-id",
                        "device_type_id": null,
                        "device_type_name": null,
                        "device_type_network": null,
                        "location_id": "location-id",
                        "name": "Lights Switch",
                        "room_id": "room-2"
                    }
                },
                "locations": {
                    "location-id": {
                        "name": "My home",
                        "location_id": "location-id",
                        "country_code": "us",
                        "locale": "en-US",
                        "latitude": "latitude",
                        "longitude": "longitude",
                        "temperature_scale": null,
                        "region_radius": null,
                        "timezone_id": null,
                        "rooms": {
                            "room-1": {
                                "background_image": null,
                                "location_id": "location-1",
                                "name": "Living Room",
                                "room_id": "room-1"
                            },
                            "room-2": {
                                "background_image": null,
                                "location_id": "location-1",
                                "name": "Bedroom",
                                "room_id": "room-2"
                            }
                        }
                    }
                }
            }

        """
        self.refresh_info()
        return {
            'locations': {
                loc.location_id: self._location_to_dict(loc) for loc in self._locations
            },
            'devices': {
                dev.device_id: self._device_to_dict(dev) for dev in self._devices
            },
        }

    @action
    def get_location(
        self, location_id: Optional[str] = None, name: Optional[str] = None
    ) -> dict:
        """
        Get the info of a location by ID or name.

          .. code-block:: json

            {
                "name": "My home",
                "location_id": "location-id",
                "country_code": "us",
                "locale": "en-US",
                "latitude": "latitude",
                "longitude": "longitude",
                "temperature_scale": null,
                "region_radius": null,
                "timezone_id": null,
                "rooms": {
                    "room-1": {
                        "background_image": null,
                        "location_id": "location-1",
                        "name": "Living Room",
                        "room_id": "room-1"
                    },
                    "room-2": {
                        "background_image": null,
                        "location_id": "location-1",
                        "name": "Bedroom",
                        "room_id": "room-2"
                    }
                }
            }

        """
        assert location_id or name, 'Specify either location_id or name'
        if (
            location_id not in self._locations_by_id
            or name not in self._locations_by_name
        ):
            self.refresh_info()

        location: Optional[Dict[str, Any]] = {}
        if location_id:
            location = self._locations_by_id.get(location_id)
        elif name:
            location = self._locations_by_name.get(name)

        assert location, f'Location {location_id or name} not found'
        return self._location_to_dict(location)

    def _get_device(self, device: str) -> DeviceEntity:
        return self._get_devices(device)[0]

    @staticmethod
    def _to_device_and_property(device: str) -> Tuple[str, Optional[str]]:
        tokens = device.split(':')
        if len(tokens) > 1:
            return (tokens[0], tokens[1])
        return tokens[0], None

    def _get_existing_and_missing_devices(
        self, *devices: str
    ) -> Tuple[List[DeviceEntity], List[str]]:
        # Split the external_id:type indicators and always return the parent device
        devices = tuple(self._to_device_and_property(dev)[0] for dev in devices)

        found_devs = {
            dev: self._devices_by_id.get(dev, self._devices_by_name.get(dev))
            for dev in devices
            if self._devices_by_id.get(dev, self._devices_by_name.get(dev))
        }

        missing_devs = {dev for dev in devices if dev not in found_devs}
        return list(found_devs.values()), list(missing_devs)

    def _get_devices(self, *devices: str) -> List[DeviceEntity]:
        devs, missing_devs = self._get_existing_and_missing_devices(*devices)
        if missing_devs:
            self.refresh_info()

        devs, missing_devs = self._get_existing_and_missing_devices(*devices)
        assert not missing_devs, f'Devices not found: {missing_devs}'
        return devs

    @action
    def get_device(self, device: str) -> dict:
        """
        Get a device info by ID or name.

        :param device: Device ID or name.
        :return:

          .. code-block:: json

            "tv-switch-id": {
                "capabilities": [
                    "switch",
                    "refresh",
                    "healthCheck"
                ],
                "device_id": "tv-switch-id",
                "device_type_id": null,
                "device_type_name": null,
                "device_type_network": null,
                "location_id": "location-id",
                "name": "TV Smart Switch",
                "room_id": "room-1"
            }

        """
        dev = self._get_device(device)
        return self._device_to_dict(dev)

    async def _execute(
        self,
        device_id: str,
        capability: str,
        command,
        component_id: str,
        args: Optional[list],
    ):
        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            api = SmartThings(session, self._access_token)
            device = await api.device(device_id)
            ret = await device.command(
                component_id=component_id,
                capability=capability,
                command=command,
                args=args,
            )

            assert (
                ret
            ), f'The command {capability}={command} failed on device {device_id}'

            await self._get_device_status(api, device_id, publish_entities=True)

    @action
    def execute(
        self,
        device: str,
        capability: str,
        command,
        component_id: str = 'main',
        args: Optional[list] = None,
    ):
        """
        Execute a command on a device.

        Example request to turn on a device with ``switch`` capability:

          .. code-block:: json

            {
              "type": "request",
              "action": "smartthings.execute",
              "args": {
                "device": "My Switch",
                "capability": "switch",
                "command": "on"
              }
            }

        :param device: Device ID or name.
        :param capability: Property to be read/written (see device ``capabilities`` returned from :meth:`.get_device`).
        :param command: Command to execute on the ``capability``
            (see https://smartthings.developer.samsung.com/docs/api-ref/capabilities.html).
        :param component_id: ID of the component to execute the command on (default: ``main``, i.e. the device itself).
        :param args: Command extra arguments, as a list.
        """
        dev = self._get_device(device)

        with self._execute_lock:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                loop.run_until_complete(
                    self._execute(
                        device_id=dev.device_id,
                        capability=capability,
                        command=command,
                        component_id=component_id,
                        args=args,
                    )
                )
            finally:
                loop.stop()

    @staticmethod
    def _property_to_entity_name(  # pylint: disable=redefined-builtin
        property: str,
    ) -> str:
        return ' '.join(
            [
                t[:1].upper() + t[1:]
                for t in camel_case_to_snake_case(property).split('_')
            ]
        )

    @classmethod
    def _to_entity(  # pylint: disable=redefined-builtin
        cls, device: DeviceEntity, property: str, entity_type: Type[Entity], **kwargs
    ) -> Entity:
        return entity_type(
            id=f'{device.device_id}:{property}',
            name=cls._property_to_entity_name(property),
            **kwargs,
        )

    @classmethod
    def _get_status_attr_info(cls, device: DeviceEntity, mapper: DeviceMapper) -> dict:
        status = device.status.attributes.get(mapper.attribute)
        info = {}
        if status:
            info.update(
                {
                    attr: getattr(status, attr, None)
                    for attr in ('unit', 'min', 'max')
                    if getattr(status, attr, None) is not None
                }
            )

            supported_values = mapper.values
            if isinstance(mapper.value_type, str):
                # The list of supported values is actually contained on a
                # device attribute
                try:
                    supported_values = getattr(
                        device.status, mapper.value_type, mapper.values
                    )
                except Exception:
                    pass

            if supported_values:
                info['values'] = mapper.values

        return info

    @staticmethod
    def _merge_dicts(*dicts: dict) -> dict:
        ret = {}
        for d in dicts:
            ret.update(d)
        return ret

    @classmethod
    def _get_supported_entities(
        cls,
        device: DeviceEntity,
        entity_type: Optional[Type[Entity]] = None,
        entity_value_attr: str = 'value',
        **default_entity_args,
    ) -> List[Entity]:
        mappers = [
            mapper
            for mapper in device_mappers
            if (entity_type is None or issubclass(mapper.entity_type, entity_type))
            and mapper.capability in device.capabilities
        ]

        return [
            cls._to_entity(
                device,
                property=mapper.attribute,
                entity_type=mapper.entity_type,
                **cls._merge_dicts(
                    {entity_value_attr: mapper.get_value(device)},
                    default_entity_args,
                    mapper.entity_args,
                    cls._get_status_attr_info(device, mapper),
                ),
            )
            for mapper in mappers
        ]

    @classmethod
    def _get_lights(cls, device: DeviceEntity) -> Iterable[Light]:
        if not (
            {Capability.color_control, Capability.color_temperature}.intersection(
                device.capabilities
            )
        ):
            return []

        light_attrs = {}
        status = device.status

        if Capability.switch in device.capabilities:
            light_attrs['on'] = status.switch
        if Capability.switch_level in device.capabilities:
            light_attrs['brightness'] = status.level
            light_attrs['brightness_min'] = 0
            light_attrs['brightness_max'] = 100
        if Capability.color_temperature in device.capabilities:
            light_attrs['temperature'] = status.color_temperature
            light_attrs['temperature_min'] = 1
            light_attrs['temperature_max'] = 30000
        if getattr(status, 'hue', None) is not None:
            light_attrs['hue'] = status.hue
            light_attrs['hue_min'] = 0
            light_attrs['hue_max'] = 100
        if getattr(status, 'saturation', None) is not None:
            light_attrs['saturation'] = status.saturation
            light_attrs['saturation_min'] = 0
            light_attrs['saturation_max'] = 100

        return [cls._to_entity(device, 'light', Light, **light_attrs)]

    @classmethod
    def _get_switches(cls, device: DeviceEntity) -> Iterable[Switch]:
        return cls._get_supported_entities(device, Switch, entity_value_attr='state')

    @classmethod
    def _get_enum_switches(cls, device: DeviceEntity) -> Iterable[Switch]:
        return cls._get_supported_entities(device, EnumSwitch)

    @classmethod
    def _get_dimmers(cls, device: DeviceEntity) -> Iterable[Dimmer]:
        return cls._get_supported_entities(device, Dimmer, min=0, max=100)

    @classmethod
    def _get_sensors(cls, device) -> Iterable[Sensor]:
        return cls._get_supported_entities(device, Sensor)

    def transform_entities(self, entities):
        compatible_entities = []

        for entity in entities:
            device_entities = [
                *self._get_lights(entity),
                *self._get_switches(entity),
                *self._get_dimmers(entity),
                *self._get_sensors(entity),
            ]

            if device_entities:
                parent = Device(
                    id=entity.device_id,
                    name=entity.label,
                )

                for child in device_entities:
                    child.parent = parent

                device_entities.insert(0, parent)

            compatible_entities += device_entities

        self._entities_by_id.update({e.id: e for e in compatible_entities})

        return super().transform_entities(compatible_entities)

    async def _get_device_status(
        self, api, device_id: str, publish_entities: bool
    ) -> dict:
        device = await api.device(device_id)
        assert device, f'No such device: {device_id}'
        await device.status.refresh()
        if publish_entities:
            self.publish_entities([device])

        self._devices_by_id[device_id] = device
        self._devices_by_name[device.label] = device
        for i, dev in enumerate(self._devices):
            if dev.device_id == device_id:
                self._devices[i] = device
                break

        return {
            'device_id': device_id,
            'name': device.label,
            **{
                cap: getattr(device.status, cap)
                for cap in device.capabilities
                if hasattr(device.status, cap)
                and not callable(getattr(device.status, cap))
            },
        }

    async def _refresh_status(
        self, devices: List[str], publish_entities: bool = True
    ) -> List[dict]:
        device_ids = []
        missing_device_ids = set()

        def parse_device_id(device):
            device_id = None
            if device in self._devices_by_id:
                device_id = device
                device_ids.append(device_id)
            elif device in self._devices_by_name:
                device_id = self._devices_by_name[device].device_id
                device_ids.append(device_id)
            else:
                missing_device_ids.add(device)

            if device_id and device in missing_device_ids:
                missing_device_ids.remove(device)

        for dev in devices:
            parse_device_id(dev)

        # Fail if some devices haven't been found after refreshing
        assert (
            not missing_device_ids
        ), f'Could not find the following devices: {list(missing_device_ids)}'

        async with aiohttp.ClientSession(timeout=self._timeout) as session:
            api = SmartThings(session, self._access_token)
            status_tasks = [
                asyncio.ensure_future(
                    self._get_device_status(
                        api, device_id, publish_entities=publish_entities
                    )
                )
                for device_id in device_ids
            ]

            return await asyncio.gather(*status_tasks)

    @action
    def status(  # pylint: disable=arguments-differ
        self, device: Optional[Union[str, List[str]]] = None, publish_entities=True, **_
    ) -> List[dict]:
        """
        Refresh and return the status of one or more devices.

        :param device: Device or list of devices to refresh (default: all)
        :return: A list containing on entry per device, and each entry containing the current device state. Example:

          .. code-block:: json

            [
              {
                "device_id": "switch-1",
                "name": "Fan",
                "switch": false
              },
              {
                "device_id": "tv-1",
                "name": "Samsung Smart TV",
                "switch": true
              }
            ]

        """
        self.refresh_info()

        if not device:
            self.refresh_info()
            devices = list(self._devices_by_id.keys())
        elif isinstance(device, str):
            devices = [device]
        else:
            devices = device

        with self._refresh_lock:
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                return loop.run_until_complete(
                    self._refresh_status(
                        list(devices), publish_entities=publish_entities
                    )
                )
            finally:
                loop.stop()

    def _set_switch(self, device: str, value: Optional[bool] = None):
        (
            device,
            property,  # pylint: disable=redefined-builtin
        ) = self._to_device_and_property(device)

        if not property:
            property = Attribute.switch

        if value is None:
            # Toggle case
            dev = self._get_device(device)
            if property == 'light':
                property = 'switch'
            else:
                assert property, 'No property specified'
                assert hasattr(
                    dev.status, property
                ), f'No such property on device "{dev.label}": "{property}"'

            value = getattr(dev.status, property, None)
            assert value is not None, (
                f'Could not get the current value of "{property}" for the '
                f'device "{dev.device_id}"'
            )

            value = not value  # Toggle
            device = dev.device_id

        return self.set_value(device, property, value)

    @action
    def on(self, device: str, *_, **__):  # pylint: disable=arguments-differ
        """
        Turn on a device with ``switch`` capability.

        :param device: Device name or ID.
        """
        return self._set_switch(device, True)

    @action
    def off(self, device: str, *_, **__):  # pylint: disable=arguments-differ
        """
        Turn off a device with ``switch`` capability.

        :param device: Device name or ID.
        """
        return self._set_switch(device, False)

    @action
    def toggle(self, device: str, *_, **__):  # pylint: disable=arguments-differ
        """
        Toggle a device with ``switch`` capability.

        :param device: Device name or ID.
        :return: Device status
        """
        return self._set_switch(device)

    @action
    def set_level(self, device: str, level: int, **kwargs):
        """
        Set the level of a device with ``switchLevel`` capabilities (e.g. the
        brightness of a lightbulb or the speed of a fan).

        :param device: Device ID or name.
        :param level: Level, usually a percentage value between 0 and 1.
        :param kwargs: Extra arguments that should be passed to :meth:`.execute`.
        """
        return self.set_value(device, Capability.switch_level, level, **kwargs)

    def _set_value(  # pylint: disable=redefined-builtin
        self, device: str, property: Optional[str] = None, value: Any = None, **kwargs
    ):
        if not property:
            device, property = self._to_device_and_property(device)

        assert property, 'No property name specified'
        assert value is not None, 'No value specified'
        entity_id = f'{device}:{property}'
        entity = self._entities_by_id.get(entity_id)
        assert entity, f'No such entity ID: {entity_id}'

        mapper = next(
            iter([m for m in device_mappers if m.attribute == property]), None
        )

        assert (
            mapper
        ), f'No mappers found to set {property}={value} on device "{device}"'
        assert (
            mapper.set_command
        ), f'The property "{property}" on the device "{device}" cannot be set'

        command = (
            mapper.set_command(value)
            if callable(mapper.set_command)
            else mapper.set_command
        )

        self.execute(
            device,
            mapper.capability,
            command,
            args=mapper.set_value_args(value),  # type: ignore
            **kwargs,
        )

        return self.status(device)

    @action
    def set(self, entity: str, value: Any, attribute: Optional[str] = None, **kwargs):
        return self.set_value(entity, property=attribute, value=value, **kwargs)

    @action
    def set_value(  # pylint: disable=redefined-builtin
        self, device: str, property: Optional[str] = None, value=None, **kwargs
    ):
        """
        Set the value of a device. It is compatible with the generic
        ``set_value`` method required by entities.

        :param device: Device ID or device+property name string in the format
            ``device_id:property``.
        :param property: Name of the property to be set. If not specified here
            then it should be specified on the ``device`` level.
        :param value: Value to set.
        """
        assert device, 'No device specified'
        try:
            return self._set_value(device, property, value, **kwargs)
        except Exception as e:
            self.logger.exception(e)
            raise AssertionError(e) from e

    @action
    def set_lights(  # pylint: disable=arguments-differ,redefined-builtin
        self,
        lights: Iterable[str],
        on: Optional[bool] = None,
        brightness: Optional[int] = None,
        hue: Optional[int] = None,
        saturation: Optional[int] = None,
        hex: Optional[str] = None,
        temperature: Optional[int] = None,
        **_,
    ):
        err = None

        with self._execute_lock:
            for light in lights:
                try:
                    if on is not None:
                        self.execute(
                            light, Capability.switch, Command.on if on else Command.off
                        )
                    if brightness is not None:
                        self.execute(
                            light,
                            Capability.switch_level,
                            Command.set_level,
                            args=[brightness],
                        )
                    if hue is not None:
                        self.execute(
                            light, Capability.color_control, Command.set_hue, args=[hue]
                        )
                    if saturation is not None:
                        self.execute(
                            light,
                            Capability.color_control,
                            Command.set_saturation,
                            args=[saturation],
                        )
                    if temperature is not None:
                        self.execute(
                            light,
                            Capability.color_temperature,
                            Command.set_color_temperature,
                            args=[temperature],
                        )
                    if hex is not None:
                        self.execute(
                            light,
                            Capability.color_control,
                            Command.set_color,
                            args=[hex],
                        )
                except Exception as e:
                    self.logger.error('Could not set attributes on %s: %s', light, e)
                    err = e

            if err:
                raise err

    @staticmethod
    def _device_status_to_dict(status: DeviceStatus) -> dict:
        status_dict = {}
        for attr in status.attributes:
            attr = camel_case_to_snake_case(attr)
            try:
                if hasattr(status, attr):
                    status_dict[attr] = getattr(status, attr)
            except Exception:
                # Ignore exceptions if retrieving status attributes that don't
                # apply to this device
                continue

        return status_dict

    def _get_devices_status_dict(self) -> Dict[str, dict]:
        return dict(
            filter(
                lambda d: bool(d[1]),
                [
                    (device_id, self._device_status_to_dict(device.status))
                    for device_id, device in self._devices_by_id.items()
                ],
            )
        )

    @staticmethod
    def _has_status_changed(status: dict, new_status: dict) -> bool:
        if not status and new_status:
            return True

        for attr, value in status.items():
            if attr in new_status:
                new_value = new_status[attr]
                if value != new_value:
                    return True

        return False

    def main(self):
        def refresh_status_safe():
            try:
                return self.status(publish_entities=False)
            except Exception as e:
                self.logger.exception(e)
                self.logger.error('Could not refresh the status: %s', e)
                self.wait_stop(3 * (self.poll_interval or 5))
                return None

        while not self.should_stop():
            updated_devices = {}
            devices = self._get_devices_status_dict()
            status = refresh_status_safe()
            if not status:
                continue

            new_devices = self._get_devices_status_dict()

            updated_devices = {
                device_id: self._devices_by_id[device_id]
                for device_id, new_status in new_devices.items()
                if self._has_status_changed(devices.get(device_id, {}), new_status)
            }

            self.publish_entities(updated_devices.values())
            devices.update(new_devices)
            self.wait_stop(self.poll_interval)
            refresh_status_safe()


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from time import time
from typing import Optional

from ._intent import Intent


@dataclass
class ConversationContext:
    """
    Context of the conversation process.
    """

    transcript: str = ''
    is_final: bool = False
    intent: Optional[Intent] = None
    timeout: Optional[float] = None
    t_start: Optional[float] = None

    def start(self):
        self.reset()
        self.t_start = time()

    def reset(self):
        self.transcript = ''
        self.intent = None
        self.is_final = False
        self.t_start = None

    @property
    def timed_out(self):
        return (
            (
                (not self.transcript and not self.is_final)
                or (not self.intent and not self.is_final)
            )
            and self.timeout
            and self.t_start
            and time() - self.t_start > self.timeout
        ) or (
            (
                (self.transcript and not self.is_final)
                or (self.intent and not self.is_final)
            )
            and self.timeout
            and self.t_start
            and time() - self.t_start > self.timeout * 2
        )


# vim:sw=4:ts=4:et:



import logging
import time
from typing import Callable, Optional

from pushbullet import Listener as _Listener


class Listener(_Listener):
    """
    Extends the Pushbullet Listener object by adding ``on_open`` and ``on_close`` handlers.
    """

    def __init__(
        self,
        *args,
        on_open: Optional[Callable[[], None]] = None,
        on_close: Optional[Callable[[], None]] = None,
        **kwargs,
    ):
        kwargs['on_error'] = self._err_callback()
        super().__init__(*args, **kwargs)
        self._on_open_hndl = on_open
        self._on_close_hndl = on_close
        self.logger = logging.getLogger(__name__)

    def _on_open(self):
        def callback(*_):
            self.connected = True
            self.last_update = time.time()
            if self._on_open_hndl:
                self._on_open_hndl()

        return callback

    def _on_close(self):
        def callback(*_):
            self.connected = False
            if self._on_close_hndl:
                try:
                    self._on_close_hndl()
                except Exception as e:
                    self.logger.warning('Pushbullet listener close error: %s', e)

        return callback

    def _err_callback(self):
        def callback(e):
            self.logger.error('Pushbullet listener error: %s: %s', type(e).__name__, e)

        return callback


# vim:sw=4:ts=4:et:



from dataclasses import dataclass
from typing import Optional, Type

from .._serialize import Serializable
from .._utils import type_str


@dataclass
class ReturnValue(Serializable):
    """
    Represents the return value of an action.
    """

    doc: Optional[str] = None
    type: Optional[Type] = None

    def to_dict(self) -> dict:
        return {
            "doc": self.doc,
            "type": type_str(self.type),
        }



import multiprocessing
import threading
from queue import Empty
from typing import Dict, List, Optional

from platypush.plugins import RunnablePlugin, action
from platypush.schemas.joystick import JoystickStatusSchema

from ._inputs import DeviceManager, GamePad
from ._manager import JoystickManager


class JoystickPlugin(RunnablePlugin):
    """
    A plugin to monitor joystick events.
    """

    def __init__(self, poll_interval: float = 0.025, **kwargs):
        """
        :param poll_interval: Polling interval in seconds (default: 0.025)
        """

        super().__init__(poll_interval=poll_interval, **kwargs)
        self._dev_managers: Dict[str, JoystickManager] = {}
        self._state_queue = multiprocessing.Queue()
        self._state_monitor_thread: Optional[threading.Thread] = None

    @staticmethod
    def _key(device: GamePad) -> str:
        return device.get_char_device_path()

    def _start_manager(self, device: GamePad):
        if device in self._dev_managers:
            return

        dev_manager = JoystickManager(
            device=device,
            poll_interval=self.poll_interval,
            state_queue=self._state_queue,
        )

        dev_manager.start()
        self._dev_managers[self._key(device)] = dev_manager

    def _stop_manager(self, device: str):
        dev_manager = self._dev_managers.get(device)
        if not dev_manager:
            return

        dev_manager.stop()
        dev_manager.join(1)
        if dev_manager.is_alive():
            dev_manager.kill()

        del self._dev_managers[device]

    def _state_monitor(self):
        while not self.should_stop():
            try:
                state = self._state_queue.get(timeout=0.5)
                device = state.device
                if device not in self._dev_managers:
                    continue

                self._dev_managers[device].state = state.state
            except Empty:
                pass
            except Exception as e:
                self.logger.exception(e)

    def main(self):
        if not self._state_monitor_thread:
            self._state_monitor_thread = threading.Thread(
                target=self._state_monitor, daemon=True
            )
            self._state_monitor_thread.start()

        while not self.should_stop():
            try:
                devices = DeviceManager().gamepads
                missing_devices_keys = set(self._dev_managers.keys()) - {
                    self._key(dev) for dev in devices
                }

                new_devices = {
                    dev for dev in devices if self._key(dev) not in self._dev_managers
                }

                # Stop managers for devices that are no longer connected
                for dev in missing_devices_keys:
                    self._stop_manager(dev)

                # Start managers for new devices
                for dev in new_devices:
                    self._start_manager(dev)
            except Exception as e:
                self.logger.exception(e)
            finally:
                self.wait_stop(max(0.5, min(10, (self.poll_interval or 0) * 10)))

    def stop(self):
        for dev in list(self._dev_managers.keys()):
            self._stop_manager(dev)

        super().stop()

    @action
    def status(self) -> List[dict]:
        """
        :return: .. schema:: joystick.JoystickStatusSchema(many=True)
        """
        return JoystickStatusSchema().dump(
            [
                {
                    'device': dev.device,
                    'state': dev.state,
                }
                for dev in self._dev_managers.values()
            ],
            many=True,
        )


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
import importlib
import inspect
import json
import logging
import os
import pathlib
import re
import shutil
import subprocess
import sys

from typing import (
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    Iterable,
    Mapping,
    Sequence,
    Set,
    Type,
    Union,
)

from platypush.utils import get_src_root, is_root

_available_package_manager = None
logger = logging.getLogger(__name__)


class BaseImage(Enum):
    """
    Supported base images for Dockerfiles.
    """

    ALPINE = 'alpine'
    DEBIAN = 'debian'
    FEDORA = 'fedora'
    UBUNTU = 'ubuntu'

    def __str__(self) -> str:
        """
        Explicit __str__ override for argparse purposes.
        """
        return self.value


@dataclass
class OSMeta:
    """
    Operating system metadata.
    """

    name: str
    description: str


class OS(Enum):
    """
    Supported operating systems.
    """

    ALPINE = OSMeta('alpine', 'Alpine')
    ARCH = OSMeta('arch', 'Arch Linux')
    DEBIAN = OSMeta('debian', 'Debian')
    FEDORA = OSMeta('fedora', 'Fedora')
    UBUNTU = OSMeta('ubuntu', 'Ubuntu')


@dataclass
class PackageManager:
    """
    Representation of a package manager.
    """

    executable: str
    """ The executable name. """
    default_os: OS
    """
    The default distro whose configuration we should use if this package
    manager is detected.
    """
    install_doc: str
    """ The base install command that will be used in the generated documentation. """
    install: Sequence[str] = field(default_factory=tuple)
    """ The install command, as a sequence of strings. """
    uninstall: Sequence[str] = field(default_factory=tuple)
    """ The uninstall command, as a sequence of strings. """
    list: Sequence[str] = field(default_factory=tuple)
    """ The command to list the installed packages. """
    parse_list_line: Callable[[str], str] = field(default_factory=lambda: lambda s: s)
    """
    Internal package-manager dependent function that parses the base package
    name from a line returned by the list command.
    """

    def _get_installed(self) -> Sequence[str]:
        """
        :return: The context-aware list of installed packages.
            It should only be used within the context of :meth:`.get_installed`.
        """

        if os.environ.get('DOCKER_CTX'):
            # If we're running in a Docker build context, don't run the package
            # manager to retrieve the list of installed packages, as the host
            # and guest systems have different environments.
            return ()

        return tuple(
            line.strip()
            for line in subprocess.Popen(  # pylint: disable=consider-using-with
                self.list, stdout=subprocess.PIPE
            )
            .communicate()[0]
            .decode()
            .split('\n')
            if line.strip()
        )

    def get_installed(self) -> Sequence[str]:
        """
        :return: The list of installed packages.
        """
        return tuple(self.parse_list_line(line) for line in self._get_installed())


class PackageManagers(Enum):
    """
    Supported package managers.
    """

    APK = PackageManager(
        executable='apk',
        install_doc='apk add',
        install=('apk', 'add', '--update', '--no-interactive', '--no-cache'),
        uninstall=('apk', 'del', '--no-interactive'),
        list=('apk', 'list', '--installed'),
        default_os=OS.ALPINE,
        parse_list_line=lambda line: re.sub(r'.*\s*\{(.+?)}\s*.*', r'\1', line),
    )

    APT = PackageManager(
        executable='apt',
        install_doc='apt install',
        install=('apt', 'install', '-y'),
        uninstall=('apt', 'remove', '-y'),
        list=('apt', 'list', '--installed'),
        default_os=OS.DEBIAN,
        parse_list_line=lambda line: line.split('/')[0],
    )

    DNF = PackageManager(
        executable='dnf',
        install_doc='yum install',
        install=('dnf', 'install', '-y'),
        uninstall=('dnf', 'remove', '-y'),
        list=('dnf', 'list', '--installed'),
        default_os=OS.FEDORA,
        parse_list_line=lambda line: re.split(r'\s+', line)[0].split('.')[0],
    )

    PACMAN = PackageManager(
        executable='pacman',
        install_doc='pacman -S',
        install=('pacman', '-S', '--noconfirm', '--needed'),
        uninstall=('pacman', '-R', '--noconfirm'),
        list=('pacman', '-Q'),
        default_os=OS.ARCH,
        parse_list_line=lambda line: line.split(' ')[0],
    )

    @classmethod
    def by_executable(cls, name: str) -> "PackageManagers":
        """
        :param name: The name of the package manager executable to get the
            package manager for.
        :return: The `PackageManager` object for the given executable.
        """
        pkg_manager = next(iter(pm for pm in cls if pm.value.executable == name), None)
        if not pkg_manager:
            raise ValueError(f'Unknown package manager: {name}')

        return pkg_manager

    @classmethod
    def get_command(cls, name: str) -> Iterable[str]:
        """
        :param name: The name of the package manager executable to get the
            command for.
        :return: The base command to execute, as a sequence of strings.
        """
        pkg_manager = next(iter(pm for pm in cls if pm.value.executable == name), None)
        if not pkg_manager:
            raise ValueError(f'Unknown package manager: {name}')

        return pkg_manager.value.install

    @classmethod
    def scan(cls) -> Optional["PackageManagers"]:
        """
        Get the name of the available package manager on the system, if supported.
        """
        # pylint: disable=global-statement
        global _available_package_manager
        if _available_package_manager:
            return _available_package_manager

        available_package_managers = [
            pkg_manager
            for pkg_manager in cls
            if shutil.which(pkg_manager.value.executable)
        ]

        if not available_package_managers:
            logger.warning(
                '\nYour OS does not provide any of the supported package managers.\n'
                'You may have to install some optional dependencies manually.\n'
                'Supported package managers: %s.\n',
                ', '.join([pm.value.executable for pm in cls]),
            )

            return None

        _available_package_manager = available_package_managers[0]
        return _available_package_manager


class InstallContext(Enum):
    """
    Supported installation contexts.
    """

    NONE = None
    DOCKER = 'docker'
    VENV = 'venv'


class ManifestType(Enum):
    """
    Manifest types.
    """

    PLUGIN = 'plugin'
    BACKEND = 'backend'


@dataclass
class Dependencies:
    """
    Dependencies for a plugin/backend.
    """

    before: List[str] = field(default_factory=list)
    """ Commands to execute before the component is installed. """
    packages: Set[str] = field(default_factory=set)
    """ System packages required by the component. """
    pip: Set[str] = field(default_factory=set)
    """ pip dependencies. """
    after: List[str] = field(default_factory=list)
    """ Commands to execute after the component is installed. """
    pkg_manager: Optional[PackageManagers] = None
    """ Override the default package manager detected on the system. """
    install_context: InstallContext = InstallContext.NONE
    """ The installation context - Docker, virtual environment or bare metal. """
    base_image: Optional[BaseImage] = None
    """ Base image used in case of Docker installations. """
    by_pkg_manager: Dict[PackageManagers, Set[str]] = field(default_factory=dict)
    """ All system dependencies, grouped by package manager. """

    def to_dict(self):
        return {
            'before': self.before,
            'packages': list(self.packages),
            'pip': self.pip,
            'after': self.after,
            'by_pkg_manager': {
                pkg_manager.value.executable: list(pkgs)
                for pkg_manager, pkgs in self.by_pkg_manager.items()
            },
        }

    @property
    def _is_venv(self) -> bool:
        """
        :return: True if the dependencies scanning logic is running either in a
            virtual environment or in a virtual environment preparation
            context.
        """
        return (
            self.install_context == InstallContext.VENV or sys.prefix != sys.base_prefix
        )

    @property
    def _is_docker(self) -> bool:
        """
        :return: True if the dependencies scanning logic is running either in a
            Docker environment.
        """
        return (
            self.install_context == InstallContext.DOCKER
            or bool(os.environ.get('DOCKER_CTX'))
            or os.path.isfile('/.dockerenv')
        )

    @property
    def _wants_sudo(self) -> bool:
        """
        :return: True if the system dependencies should be installed with sudo.
        """
        return not (self._is_docker or is_root())

    @staticmethod
    def _get_requirements_dir() -> str:
        """
        :return: The root folder for the base installation requirements.
        """
        return os.path.join(get_src_root(), 'install', 'requirements')

    @classmethod
    def _parse_requirements_file(
        cls,
        requirements_file: str,
        install_context: InstallContext = InstallContext.NONE,
    ) -> Iterable[str]:
        """
        :param requirements_file: The path to the requirements file.
        :return: The list of requirements to install.
        """
        with open(requirements_file, 'r') as f:
            return {
                line.strip()
                for line in f
                if not (
                    not line.strip()
                    or line.strip().startswith('#')
                    # Virtual environments will install all the Python
                    # dependencies via pip, so we should skip them here
                    or (
                        install_context == InstallContext.VENV
                        and cls._is_python_pkg(line.strip())
                    )
                )
            }

    @classmethod
    def _get_base_system_dependencies(
        cls,
        pkg_manager: Optional[PackageManagers] = None,
        install_context: InstallContext = InstallContext.NONE,
    ) -> Iterable[str]:
        """
        :return: The base system dependencies that should be installed on the
            system.
        """

        # Docker images will install the base packages through their own
        # dedicated shell script, so don't report their base system
        # requirements here.
        if not (pkg_manager and install_context != InstallContext.DOCKER):
            return set()

        return cls._parse_requirements_file(
            os.path.join(
                cls._get_requirements_dir(),
                pkg_manager.value.default_os.name.lower() + '.txt',
            ),
            install_context,
        )

    @staticmethod
    def _is_python_pkg(pkg: str) -> bool:
        """
        Utility function that returns True if a given package is a Python
        system package. These should be skipped during a virtual
        environment installation, as the virtual environment will be
        installed via pip.
        """
        tokens = pkg.split('-')
        return len(tokens) > 1 and tokens[0] in {'py3', 'python3', 'python'}

    @classmethod
    def from_config(
        cls,
        conf_file: Optional[str] = None,
        pkg_manager: Optional[PackageManagers] = None,
        install_context: InstallContext = InstallContext.NONE,
        base_image: Optional[BaseImage] = None,
    ) -> "Dependencies":
        """
        Parse the required dependencies from a configuration file.
        """
        if not pkg_manager:
            pkg_manager = PackageManagers.scan()

        base_system_deps = cls._get_base_system_dependencies(
            pkg_manager=pkg_manager, install_context=install_context
        )

        deps = cls(
            packages=set(base_system_deps),
            pkg_manager=pkg_manager,
            install_context=install_context,
            base_image=base_image,
        )

        for manifest in Manifests.by_config(conf_file, pkg_manager=pkg_manager):
            deps.before += manifest.install.before
            deps.pip.update(manifest.install.pip)
            deps.packages.update(manifest.install.packages)
            deps.after += manifest.install.after

        return deps

    def to_pkg_install_commands(self) -> Generator[str, None, None]:
        """
        Generates the package manager commands required to install the given
        dependencies on the system.
        """

        pkg_manager = self.pkg_manager or PackageManagers.scan()

        if self.packages and pkg_manager:
            installed_packages = pkg_manager.value.get_installed()
            to_install = sorted(
                pkg
                for pkg in self.packages  # type: ignore
                if pkg not in installed_packages
                and not (
                    self.install_context == InstallContext.VENV
                    and self._is_python_pkg(pkg)
                )
            )

            if to_install:
                yield ' '.join(
                    [
                        *(['sudo'] if self._wants_sudo else []),
                        *pkg_manager.value.install,
                        '\\\n  '.join(to_install),
                    ]
                )

    def to_pip_install_commands(self, full_command=True) -> Generator[str, None, None]:
        """
        Generates the pip commands required to install the given dependencies on
        the system.

        :param full_command: Whether to return the full pip command to execute
            (as a single string) or the list of packages that will be installed
            through another script.
        """
        wants_break_system_packages = (
            # --break-system-packages has been introduced in Python 3.10
            sys.version_info >= (3, 11)
            # If we're generating a Docker image then we always need
            # --break-system-packages
            or self._is_docker
        )

        if self._is_venv and not self._is_docker:
            # If we're in a virtual environment then we don't need
            # --break-system-packages, unless we're generating a Docker image
            wants_break_system_packages = False

        if self.pip:
            deps = sorted(self.pip)
            if full_command:
                yield (
                    'pip install -U --no-input '
                    + ('--no-cache-dir ' if self._is_docker else '')
                    + (
                        '--break-system-packages '
                        if wants_break_system_packages
                        else ''
                    )
                    + ' \\\n  '.join(deps)
                )
            else:
                yield from deps

    def to_install_commands(self) -> Generator[str, None, None]:
        """
        Generates the commands required to install the given dependencies on
        this system.
        """
        for cmd in self.before:
            yield cmd

        for cmd in self.to_pkg_install_commands():
            yield cmd

        for cmd in self.to_pip_install_commands():
            yield cmd

        for cmd in self.after:
            yield cmd


class Manifest(ABC):
    """
    Base class for plugin/backend manifests.
    """

    def __init__(
        self,
        package: str,
        description: Optional[str] = None,
        install: Optional[Dict[str, Iterable[str]]] = None,
        events: Optional[Mapping] = None,
        pkg_manager: Optional[PackageManagers] = None,
        **_,
    ):
        self._pkg_manager = pkg_manager or PackageManagers.scan()
        self.description = description
        self.install = self._init_deps(install or {})
        self.events = self._init_events(events or {})
        self.package = package
        self.component_name = '.'.join(package.split('.')[2:])
        self.component = None

    @property
    @abstractmethod
    def manifest_type(self) -> ManifestType:
        """
        :return: The type of the manifest.
        """

    @property
    def file(self) -> str:
        """
        :return: The path to the manifest file.
        """
        return os.path.join(
            get_src_root(),
            *self.package.split('.')[1:],
            'manifest.json',
        )

    def _init_deps(self, install: Mapping[str, Iterable[str]]) -> Dependencies:
        deps = Dependencies()
        for key, items in install.items():
            if key == 'pip':
                deps.pip.update(items)
            elif key == 'before':
                deps.before += items
            elif key == 'after':
                deps.after += items
            else:
                deps.by_pkg_manager[PackageManagers.by_executable(key)] = set(items)
                if self._pkg_manager and key == self._pkg_manager.value.executable:
                    deps.packages.update(items)

        return deps

    @staticmethod
    def _init_events(
        events: Union[Iterable[str], Mapping[str, Optional[str]]]
    ) -> Dict[Type, str]:
        evt_dict = events if isinstance(events, Mapping) else {e: None for e in events}
        ret = {}

        for evt_name, doc in evt_dict.items():
            evt_module_name, evt_class_name = evt_name.rsplit('.', 1)
            try:
                evt_module = importlib.import_module(evt_module_name)
                evt_class = getattr(evt_module, evt_class_name)
            except Exception as e:
                raise AssertionError(f'Could not load event {evt_name}: {e}') from e

            ret[evt_class] = doc or evt_class.__doc__

        return ret

    @classmethod
    def from_file(
        cls, filename: str, pkg_manager: Optional[PackageManagers] = None
    ) -> "Manifest":
        """
        Parse a manifest filename into a ``Manifest`` class.
        """
        with open(str(filename), 'r') as f:
            manifest = json.load(f).get('manifest', {})

        assert 'type' in manifest, f'Manifest file {filename} has no type field'
        comp_type = ManifestType(manifest.pop('type'))
        manifest_class = cls.by_type(comp_type)
        return manifest_class(**manifest, pkg_manager=pkg_manager)

    @classmethod
    def by_type(cls, manifest_type: ManifestType) -> Type["Manifest"]:
        """
        :return: The manifest class corresponding to the given manifest type.
        """
        if manifest_type == ManifestType.PLUGIN:
            return PluginManifest
        if manifest_type == ManifestType.BACKEND:
            return BackendManifest

        raise ValueError(f'Unknown manifest type: {manifest_type}')

    def __repr__(self):
        """
        :return: A JSON serialized representation of the manifest.
        """
        return json.dumps(
            {
                'description': self.description,
                'install': self.install,
                'events': {
                    '.'.join([evt_type.__module__, evt_type.__name__]): doc
                    for evt_type, doc in self.events.items()
                },
                'type': self.manifest_type.value,
                'package': self.package,
                'component_name': self.component_name,
            }
        )


class PluginManifest(Manifest):
    """
    Plugin manifest.
    """

    @property
    def manifest_type(self) -> ManifestType:
        return ManifestType.PLUGIN


# pylint: disable=too-few-public-methods
class BackendManifest(Manifest):
    """
    Backend manifest.
    """

    @property
    def manifest_type(self) -> ManifestType:
        return ManifestType.BACKEND


class Manifests:
    """
    General-purpose manifests utilities.
    """

    @staticmethod
    def by_base_class(
        base_class: Type, pkg_manager: Optional[PackageManagers] = None
    ) -> Generator[Manifest, None, None]:
        """
        Get all the manifest files declared under the base path of a given class
        and parse them into :class:`Manifest` objects.
        """
        for mf in pathlib.Path(os.path.dirname(inspect.getfile(base_class))).rglob(
            'manifest.json'
        ):
            try:
                yield Manifest.from_file(str(mf), pkg_manager=pkg_manager)
            except Exception as e:
                logger.debug(
                    'Could not parse manifest file %s: %s', mf, e, exc_info=True
                )

    @staticmethod
    def by_config(
        conf_file: Optional[str] = None,
        pkg_manager: Optional[PackageManagers] = None,
    ) -> Generator[Manifest, None, None]:
        """
        Get all the manifest objects associated to the extensions declared in a
        given configuration file.
        """
        from platypush.config import Config

        conf_args = []
        if conf_file:
            conf_args.append(conf_file)

        Config.init(*conf_args)
        app_dir = get_src_root()

        for name in Config.get_backends().keys():
            yield Manifest.from_file(
                os.path.join(app_dir, 'backend', *name.split('.'), 'manifest.json'),
                pkg_manager=pkg_manager,
            )

        for name in Config.get_plugins().keys():
            yield Manifest.from_file(
                os.path.join(app_dir, 'plugins', *name.split('.'), 'manifest.json'),
                pkg_manager=pkg_manager,
            )

    @staticmethod
    def scan() -> Generator[Manifest, None, None]:
        """
        Scan all the manifest files in the source tree and parse them into
        :class:`Manifest` objects.
        """
        for mf in pathlib.Path(get_src_root()).rglob('manifest.json'):
            yield Manifest.from_file(str(mf))



import datetime

from sqlalchemy import (
    Float,
    Column,
    Integer,
    String,
    DateTime,
    PrimaryKeyConstraint,
    ForeignKey,
)
from sqlalchemy.orm import sessionmaker, scoped_session

from platypush.common.db import declarative_base

Base = declarative_base()
Session = scoped_session(sessionmaker())


class MediaDirectory(Base):
    """Models the MediaDirectory table"""

    __tablename__ = 'MediaDirectory'
    __table_args__ = {'sqlite_autoincrement': True}

    id = Column(Integer, primary_key=True)
    path = Column(String)
    last_indexed_at = Column(DateTime)

    @classmethod
    def build(cls, path, last_indexed_at=None, id=None):
        record = cls()
        record.id = id
        record.path = path
        record.last_indexed_at = last_indexed_at
        return record


class MediaFile(Base):
    """Models the MediaFile table"""

    __tablename__ = 'MediaFile'
    __table_args__ = {'sqlite_autoincrement': True}

    id = Column(Integer, primary_key=True)
    directory_id = Column(
        Integer, ForeignKey('MediaDirectory.id', ondelete='CASCADE'), nullable=False
    )
    path = Column(String, nullable=False, unique=True)
    duration = Column(Float)
    width = Column(Integer)
    height = Column(Integer)
    image = Column(String)
    created_at = Column(DateTime)
    indexed_at = Column(DateTime)

    @classmethod
    def build(cls, directory_id, path, indexed_at=None, id=None, **kwargs):
        record = cls()
        record.id = id
        record.directory_id = directory_id
        record.path = path
        record.indexed_at = indexed_at or datetime.datetime.now()

        for k, v in kwargs.items():
            setattr(record, k, v)

        return record


class MediaToken(Base):
    """Models the MediaToken table"""

    __tablename__ = 'MediaToken'
    __table_args__ = {'sqlite_autoincrement': True}

    id = Column(Integer, primary_key=True)
    token = Column(String, nullable=False, unique=True)

    @classmethod
    def build(cls, token, id=None):
        record = cls()
        record.id = id
        record.token = token
        return record


class MediaFileToken(Base):
    """Models the MediaFileToken table"""

    __tablename__ = 'MediaFileToken'

    file_id = Column(
        Integer, ForeignKey('MediaFile.id', ondelete='CASCADE'), nullable=False
    )
    token_id = Column(
        Integer, ForeignKey('MediaToken.id', ondelete='CASCADE'), nullable=False
    )

    __table_args__ = (PrimaryKeyConstraint(file_id, token_id), {})

    @classmethod
    def build(cls, file_id, token_id, id=None):
        record = cls()
        record.id = id
        record.file_id = file_id
        record.token_id = token_id
        return record


# vim:sw=4:ts=4:et:



from functools import wraps
from threading import Event, RLock
from typing import Any, Callable, Collection, Dict, Iterable, Optional, Union

from platypush.entities import EntityManager, get_entities_engine
from platypush.entities.variables import Variable
from platypush.plugins import Plugin, action


# pylint: disable=protected-access
def ensure_initialized(f: Callable[..., Any]):
    """
    Ensures that the entities engine has been initialized before
    reading/writing the db.

    It also performs lazy initialization of the variables in the local cache.
    """

    @wraps(f)
    def wrapper(*args, **kwargs):
        self: VariablePlugin = args[0]

        if not self._initialized.is_set():
            with self._init_lock:
                get_entities_engine(timeout=20)

                if not self._initialized.is_set():
                    self._initialized.set()
                    with self._db.get_session() as session:
                        self._db_vars.update(
                            {
                                str(var.name): var.value
                                for var in session.query(Variable).all()
                            }
                        )

        return f(*args, **kwargs)

    return wrapper


class VariablePlugin(Plugin, EntityManager):
    """
    This plugin allows you to manipulate context variables that can be
    accessed across your tasks. It requires the :mod:`platypush.plugins.db`
    and :mod:`platypush.plugins.redis` plugins to be enabled, as the variables
    will be stored either persisted on a local database or on the local Redis instance.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._db_vars: Dict[str, Optional[str]] = {}
        """ Local cache for db variables. """
        self._initialized = Event()
        """ Lazy initialization event for the _db_vars map. """
        self._init_lock = RLock()
        """ Lock for the _db_vars map initialization. """

    @ensure_initialized
    @action
    def get(self, name: Optional[str] = None, default_value=None):
        """
        Get the value of a variable by name from the local db.

        :param name: Variable name. If not specified, all the stored variables will be returned.
        :param default_value: What will be returned if the variable is not defined (default: None)
        :returns: A map in the format ``{"<name>":"<value>"}``
        """

        return (
            {name: self._db_vars.get(name, default_value)}
            if name is not None
            else self.status().output
        )

    @ensure_initialized
    @action
    def set(self, **kwargs):
        """
        Set a variable or a set of variables on the local db.

        :param kwargs: Key-value list of variables to set (e.g. ``foo='bar', answer=42``)
        """

        self.publish_entities(kwargs)
        self._db_vars.update(kwargs)
        return kwargs

    @ensure_initialized
    @action
    def delete(self, name: str):
        """
        Delete a variable from the database.

        Unlike :meth:`.unset`, this method actually deletes the record from the
        database instead of setting it to null.

        :param name: Name of the variable to remove.
        """

        with self._db.get_session() as session:
            entity = session.query(Variable).filter(Variable.name == name).first()
            if entity is not None:
                self._entities.delete(entity.id)

        self._db_vars.pop(name, None)
        return True

    @ensure_initialized
    @action
    def unset(self, name: str):
        """
        Unset a variable by name if it's set on the local db.

        Unlike :meth:`.delete`, this method only sets the record to null
        instead of removing it from the database.

        :param name: Name of the variable to remove.
        """

        return self.set(**{name: None})

    @ensure_initialized
    @action
    def mget(self, name: str):
        """
        Get the value of a variable by name from Redis.

        :param name: Variable name
        :returns: A map in the format ``{"<name>":"<value>"}``
        """

        return self._redis.mget([name])

    @ensure_initialized
    @action
    def mset(self, **kwargs):
        """
        Set a variable or a set of variables on Redis.

        :param kwargs: Key-value list of variables to set (e.g. ``foo='bar', answer=42``)
        :returns: A map with the set variables
        """

        self._redis.mset(**kwargs)
        return kwargs

    @ensure_initialized
    @action
    def munset(self, name: str):
        """
        Unset a Redis variable by name if it's set

        :param name: Name of the variable to remove
        """

        return self._redis.delete(name)

    @ensure_initialized
    @action
    def expire(self, name: str, expire: int):
        """
        Set a variable expiration on Redis

        :param name: Variable name
        :param expire: Expiration time in seconds
        """

        return self._redis.expire(name, expire)

    def transform_entities(
        self, entities: Union[dict, Iterable]
    ) -> Collection[Variable]:
        variables = (
            [
                {
                    'name': name,
                    'value': value,
                }
                for name, value in entities.items()
            ]
            if isinstance(entities, dict)
            else entities
        )

        return super().transform_entities(
            [
                Variable(id=var['name'], name=var['name'], value=var['value'])
                for var in variables
            ]
        )

    @ensure_initialized
    @action
    def status(self, *_, **__):
        variables = {
            name: value for name, value in self._db_vars.items() if value is not None
        }

        self.publish_entities(variables)
        return variables


# vim:sw=4:ts=4:et:



from ._battery import Battery, BatterySchema
from ._connection import Connection, ConnectionSchema
from ._cpu import (
    Cpu,
    CpuFrequency,
    CpuFrequencySchema,
    CpuInfo,
    CpuInfoSchema,
    CpuStats,
    CpuStatsSchema,
    CpuTimes,
    CpuTimesSchema,
)
from ._disk import Disk, DiskSchema
from ._fan import Fan, FanSchema
from ._memory import MemoryStats, MemoryStatsSchema, SwapStats, SwapStatsSchema
from ._model import SystemInfo
from ._network import NetworkInterface, NetworkInterfaceSchema
from ._process import Process, ProcessSchema
from ._schemas import SystemInfoSchema
from ._temperature import Temperature, TemperatureSchema
from ._user import User, UserSchema


__all__ = [
    "Battery",
    "BatterySchema",
    "Connection",
    "ConnectionSchema",
    "Cpu",
    "CpuFrequency",
    "CpuFrequencySchema",
    "CpuInfo",
    "CpuInfoSchema",
    "CpuStats",
    "CpuStatsSchema",
    "CpuTimes",
    "CpuTimesSchema",
    "Disk",
    "DiskSchema",
    "Fan",
    "FanSchema",
    "MemoryStats",
    "MemoryStatsSchema",
    "Process",
    "ProcessSchema",
    "SwapStats",
    "SwapStatsSchema",
    "NetworkInterface",
    "NetworkInterfaceSchema",
    "SystemInfo",
    "SystemInfoSchema",
    "Temperature",
    "TemperatureSchema",
    "User",
    "UserSchema",
]



from typing import Collection, Optional

from platypush.plugins import Plugin


class GooglePlugin(Plugin):
    """
    Integrates with the Google APIs using the google-api-python-client.

    This class is extended by ``GoogleMailPlugin``, ``GoogleCalendarPlugin`` etc.

    In order to use Google services (like GMail, Maps, Calendar etc.) with
    your account you need to:

        1. Create your Google application, if you don't have one already, on
           the `developers console <https://console.developers.google.com>`_.

        2. You may have to explicitly enable your user to use the app if the app
           is created in test mode. Go to "OAuth consent screen" and add your user's
           email address to the list of authorized users.

        3. Select the scopes that you want to enable for your application, depending
           on the integrations that you want to use.
           See https://developers.google.com/identity/protocols/oauth2/scopes
           for a list of the available scopes.

        4. Click on "Credentials", then "Create credentials" -> "OAuth client ID".

        5. Select "Desktop app", enter whichever name you like, and click "Create".

        6. Click on the "Download JSON" icon next to your newly created client ID.
           Save the JSON file under
           ``<WORKDIR>/credentials/google/client_secret.json``.

        7. If you're running the service on a desktop environment, then you
           can just start the application. A browser window will open and
           you'll be asked to authorize the application - you may be prompted
           with a warning because you are running a personal and potentially
           unverified application. After authorizing the application, the
           process will save the credentials under
           ``<WORKDIR>/credentials/google/<list,of,scopes>.json`` and proceed
           with the plugin initialization.

        8. If you're running the service on a headless environment, or you
           prefer to manually generate the credentials file before copying to
           another machine, you can run the following command:

               .. code-block:: bash

                  mkdir -p <WORKDIR>/credentials/google
                  python -m platypush.plugins.google.credentials \
                      'calendar.readonly,gmail.modify' \
                      <WORKDIR>/credentials/google/client_secret.json [--noauth_local_webserver]

           When launched with ``--noauth_local_webserver``, the script will
           start a local webserver and print a URL that should be opened in
           your browser. After authorizing the application, you may be
           prompted with a code that you should copy and paste back to the
           script. Otherwise, if you're running the script on a desktop, a
           browser window will be opened automatically.

    """

    def __init__(
        self,
        scopes: Optional[Collection[str]] = None,
        secrets_path: Optional[str] = None,
        **kwargs
    ):
        """
        :param scopes: List of scopes required by the API.
            See https://developers.google.com/identity/protocols/oauth2/scopes
            for a list of the available scopes. Override it in your configuration
            only if you need specific scopes that aren't normally required by the
            plugin.
        :param secrets_path: Path to the client secrets file.
            You can create your secrets.json from https://console.developers.google.com.
            Default: ``<PLATYPUSH_WORKDIR>/credentials/google/client_secret.json``.
        """

        from platypush.plugins.google.credentials import (
            get_credentials,
            default_secrets_file,
        )

        super().__init__(**kwargs)
        self._scopes = scopes or []
        self._secrets_path: str = secrets_path or default_secrets_file

        if self._scopes:
            scopes = " ".join(sorted(self._scopes))
            self.credentials = {
                scopes: get_credentials(scopes, secrets_file=self._secrets_path)
            }
        else:
            self.credentials = {}

    def get_service(
        self, service: str, version: str, scopes: Optional[Collection[str]] = None
    ):
        import httplib2
        from apiclient import discovery

        if scopes is None:
            scopes = getattr(self, "scopes", [])

        scopes = " ".join(sorted(scopes))
        credentials = self.credentials[scopes]
        http = credentials.authorize(httplib2.Http())
        return discovery.build(service, version, http=http, cache_discovery=False)


# vim:sw=4:ts=4:et:



from logging import getLogger
from queue import Full, Queue
from threading import Thread
from time import time
from typing import Any, Callable, Optional

from ._mix import Mix


class AudioGenerator(Thread):
    """
    The ``AudioGenerator`` class is a thread that generates synthetic raw audio
    waves and dispatches them to a queue that can be consumed by other players,
    streamers and converters.
    """

    def __init__(
        self,
        *args,
        audio_queue: Queue,  # Queue[NDArray[np.number]],
        mix: Mix,
        blocksize: int,
        sample_rate: int,
        queue_timeout: Optional[float] = None,
        should_stop: Callable[[], bool] = lambda: False,
        wait_running: Callable[[], Any] = lambda: None,
        on_stop: Callable[[], Any] = lambda: None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._audio_queue = audio_queue
        self._t_start: float = 0
        self._blocksize: int = blocksize
        self._sample_rate: int = sample_rate
        self._blocktime = self._blocksize / self._sample_rate
        self._should_stop = should_stop
        self._queue_timeout = queue_timeout
        self._wait_running = wait_running
        self._on_stop = on_stop
        self.mix = mix
        self.logger = getLogger(__name__)

    def _next_t(self, t: float) -> float:
        """
        Calculates the next starting time for the wave function.
        """
        return (
            min(t + self._blocktime, self._duration)
            if self._duration is not None
            else t + self._blocktime
        )

    def should_stop(self) -> bool:
        """
        Stops if the upstream dependencies have signalled to stop or if the
        duration is set and we have reached it.
        """
        return self._should_stop() or (
            self._duration is not None and time() - self._t_start >= self._duration
        )

    @property
    def _duration(self) -> Optional[float]:
        """
        Proxy to the mix object's duration.
        """
        return self.mix.duration()

    def run(self):
        super().run()
        self._t_start = time()
        t = 0

        while not self.should_stop():
            self._wait_running()
            if self.should_stop():
                break

            next_t = self._next_t(t)

            try:
                data = self.mix.get_wave(
                    t_start=t, t_end=next_t, sample_rate=self._sample_rate
                )
            except Exception as e:
                self.logger.warning('Could not generate the audio wave: %s', e)
                break

            try:
                self._audio_queue.put(data, timeout=self._queue_timeout)
                t = next_t
            except Full:
                self.logger.warning(
                    'The processing queue is full: either the audio consumer is stuck, '
                    'or you may want to increase queue_size'
                )

        self._on_stop()



import os
import time
from typing import IO, Optional, Union

from platypush.plugins import action
from platypush.plugins.camera import CameraPlugin, Camera

from .model import PiCameraInfo, PiCamera


class CameraPiPlugin(CameraPlugin):
    """
    Plugin to interact with a `Pi Camera
    <https://www.raspberrypi.com/documentation/accessories/camera.html>`_.

    This integration is intended to work with the `picamera2
    <https://github.com/raspberrypi/picamera2>`_ module.

    If you are running a very old OS that only provides the deprecated
    `picamera <https://github.com/waveform80/picamera>`_ module, or you rely on
    features that are currently only supported by the old module, you should
    use :class:`platypush.plugins.camera.pi.legacy.CameraPiLegacyPlugin`
    instead.
    """

    _camera_class = PiCamera
    _camera_info_class = PiCameraInfo

    _awb_modes = [
        "Auto",
        "Incandescent",
        "Tungsten",
        "Fluorescent",
        "Indoor",
        "Daylight",
        "Cloudy",
    ]

    def __init__(
        self,
        device: int = 0,
        fps: float = 30.0,
        warmup_seconds: float = 2.0,
        sharpness: float = 1.0,
        contrast: float = 1.0,
        brightness: float = 0.0,
        iso: int = 0,
        exposure_compensation: float = 0.0,
        awb_mode: str = 'Auto',
        stream_format: str = 'h264',
        **camera,
    ):
        """
        :param device: Camera device number (default: 0). Only supported on
            devices with multiple camera slots.
        :param fps: Frames per second (default: 30.0).
        :param warmup_seconds: Seconds to wait for the camera to warm up
            before taking a photo (default: 2.0).
        :param sharpness: Sharpness level, as a float between 0.0 and 16.0,
            where 1.0 is the default value, and higher values are mapped to
            higher sharpness levels.
        :param contrast: Contrast level, as a float between 0.0 and 32.0, where
            1.0 is the default value, and higher values are mapped to higher
            contrast levels.
        :param brightness: Brightness level, as a float between -1.0 and 1.0.
        :param iso: ISO level (default: 0).
        :param exposure_compensation: Exposure compensation level, as a float
            between -8.0 and 8.0.
        :param awb_mode: Auto white balance mode. Allowed values:

            - ``Auto`` (default)
            - ``Daylight``
            - ``Cloudy``
            - ``Indoor``
            - ``Fluorescent``

        :param stream_format: Default format for the output when streamed to a
            network device. Available:

                - ``h264`` (default)
                - ``mjpeg``

        :param camera: Options for the base camera plugin (see
            :class:`platypush.plugins.camera.CameraPlugin`).
        """
        super().__init__(
            device=device,
            fps=fps,
            warmup_seconds=warmup_seconds,
            stream_format=stream_format,
            **camera,
        )

        self.camera_info.sharpness = sharpness  # type: ignore
        self.camera_info.contrast = contrast  # type: ignore
        self.camera_info.brightness = brightness  # type: ignore
        self.camera_info.iso = iso  # type: ignore
        self.camera_info.exposure_compensation = exposure_compensation  # type: ignore
        self.camera_info.awb_mode = awb_mode  # type: ignore

    def _get_transform(self, device: Camera):
        from libcamera import Orientation, Transform  # type: ignore
        from picamera2.utils import orientation_to_transform  # type: ignore

        rot = device.info.rotate
        if not rot:
            return Transform(
                # It may seem counterintuitive, but the picamera2 library's flip
                # definition is the opposite of ours
                hflip=device.info.vertical_flip,
                vflip=device.info.horizontal_flip,
            )

        if rot == 90:
            orient = (
                Orientation.Rotate90Mirror
                if device.info.vertical_flip
                else Orientation.Rotate90
            )
        elif rot == 180:
            orient = (
                Orientation.Rotate180Mirror
                if device.info.horizontal_flip
                else Orientation.Rotate180
            )
        elif rot == 270:
            orient = (
                Orientation.Rotate270Mirror
                if device.info.vertical_flip
                else Orientation.Rotate270
            )
        else:
            raise AssertionError(
                f'Invalid rotation: {rot}. Supported values: 0, 90, 180, 270'
            )

        return orientation_to_transform(orient)

    def prepare_device(
        self,
        device: Camera,
        start: bool = True,
        video: bool = False,
        stream: bool = False,
        **_,
    ):
        from picamera2 import Picamera2  # type: ignore

        assert isinstance(device, PiCamera), f'Invalid device type: {type(device)}'
        camera = Picamera2(camera_num=device.info.device)
        still = not (video or stream)
        cfg_params = {
            'main': {
                'format': 'XBGR8888' if not still else 'BGR888',
                **(
                    {'size': tuple(map(int, device.info.resolution))}
                    if device.info.resolution
                    else {}
                ),
            },
            **(
                {'transform': self._get_transform(device)}
                if not still
                # We don't need to flip the image for individual frames, the base camera
                # class methods will take care of that
                else {}
            ),
            'controls': {
                'Brightness': float(device.info.brightness),
                'Contrast': float(device.info.contrast),
                'Sharpness': float(device.info.sharpness),
                'AwbMode': self._awb_modes.index(device.info.awb_mode),
            },
        }

        cfg = (
            camera.create_video_configuration
            if not still
            else camera.create_still_configuration
        )(**cfg_params)

        camera.configure(cfg)
        if start:
            camera.start()
            time.sleep(max(1, device.info.warmup_seconds))

        return camera

    def release_device(self, device: Camera):
        if device.object:
            device.object.stop()
            device.object.close()

    def capture_frame(self, device: Camera, *_, **__):
        assert device.object, 'Camera not open'
        return device.object.capture_image('main')

    @property
    def _video_encoders_by_format(self) -> dict:
        from picamera2.encoders import H264Encoder, MJPEGEncoder  # type: ignore

        return {
            'h264': H264Encoder,
            'mjpeg': MJPEGEncoder,
        }

    @action
    def capture_video(
        self,
        device: Optional[int] = None,
        duration: Optional[float] = None,
        video_file: Optional[str] = None,
        preview: bool = False,
        **camera,
    ) -> Optional[Union[str, dict]]:
        """
        Capture a video.

        :param device: 0-based index of the camera to capture from, if the
            device supports multiple cameras. Default: use the configured
            camera index or the first available camera.
        :param duration: Record duration in seconds (default: None, record
            until :meth:`.stop_capture``).
        :param video_file: If set, the stream will be recorded to the specified
            video file (default: None).
        :param camera: Camera parameters override - see constructors parameters.
        :param preview: Show a preview of the camera frames.
        :return: If duration is specified, the method will wait until the
            recording is done and return the local path to the recorded
            resource. Otherwise, it will return the status of the camera device
            after starting it.
        """
        from picamera2 import Picamera2  # type: ignore
        from picamera2.encoders import H264Encoder  # type: ignore

        assert video_file, 'Video file is required'
        camera = self.open_device(
            device=device, ctx={'start': False, 'video': True}, **camera
        )

        encoder = H264Encoder()
        assert camera.object, 'Camera not open'
        assert isinstance(
            camera.object, Picamera2
        ), f'Invalid camera object type: {type(camera.object)}'

        if preview:
            camera.object.start_preview()

        # Only H264 is supported for now
        camera.object.start_recording(encoder, os.path.expanduser(video_file))

        if duration:
            self.wait_stop(duration)
            try:
                if preview:
                    camera.object.stop_preview()
            finally:
                if camera.object:
                    camera.object.stop_recording()
                    camera.object.close()

            return video_file

        return self.status(camera.info.device).output

    def _streaming_loop(self, camera: Camera, stream_format: str, sock: IO, *_, **__):
        from picamera2 import Picamera2  # type: ignore
        from picamera2.outputs import FileOutput  # type: ignore

        encoder_cls = self._video_encoders_by_format.get(stream_format.lower())
        assert (
            encoder_cls
        ), f'Invalid stream format: {stream_format}. Supported formats: {", ".join(self._video_encoders_by_format)}'
        assert isinstance(camera, PiCamera), f'Invalid camera type: {type(camera)}'
        assert camera.object and isinstance(
            camera.object, Picamera2
        ), f'Invalid camera object type: {type(camera.object)}'

        cam = camera.object
        encoder = encoder_cls()
        cam.encoders = encoder
        encoder.output = FileOutput(sock)
        cam.start_encoder(encoder)
        cam.start()

    def _prepare_stream_writer(self, *_, **__):
        """
        Overrides the base method to do nothing - the stream writer is handled
        by the picamera2 library.
        """

    def _cleanup_stream(self, camera: Camera, *_, **__):
        cam = camera.object
        if not cam:
            return

        cam.stop()
        cam.stop_encoder()
        cam.close()


# vim:sw=4:ts=4:et:



import copy
import json
import logging
import random
import re
import time

from dataclasses import dataclass, field
from datetime import date
from typing import Any

from platypush.config import Config
from platypush.message import Message
from platypush.utils import get_event_class_by_type

logger = logging.getLogger('platypush')


class Event(Message):
    """Event message class"""

    # If this class property is set to false then the logging of these events
    # will be disabled. Logging is usually disabled for events with a very
    # high frequency that would otherwise pollute the logs e.g. camera capture
    # events
    # pylint: disable=redefined-builtin
    def __init__(
        self,
        target=None,
        origin=None,
        id=None,
        timestamp=None,
        logging_level=logging.INFO,
        disable_web_clients_notification=False,
        **kwargs,
    ):
        """
        :param target: Target node
        :type target: str
        :param origin: Origin node (default: current node)
        :type origin: str
        :param id: Event ID (default: auto-generated)
        :type id: str
        :param timestamp: Event timestamp (default: current time)
        :type timestamp: float
        :param logging_level: Logging level for this event (default:
            ``logging.INFO``)
        :param disable_web_clients_notification: Don't send a notification of
            this event to the websocket clients
        :param kwargs: Additional arguments for the event
        """

        super().__init__(timestamp=timestamp, logging_level=logging_level)
        self.id = id if id else self._generate_id()
        self.target = target if target else Config.get('device_id')
        self.origin = origin if origin else Config.get('device_id')
        self.type = f'{self.__class__.__module__}.{self.__class__.__name__}'
        self.args = kwargs
        self.disable_web_clients_notification = disable_web_clients_notification
        self._logger = logging.getLogger('platypush:events')

        for arg, value in self.args.items():
            if arg not in [
                'id',
                'args',
                'origin',
                'target',
                'type',
                'timestamp',
                'logging_level',
            ] and not arg.startswith('_'):
                self.__setattr__(arg, value)

    @classmethod
    def build(cls, msg):
        """
        Builds an event message from a JSON UTF-8 string/bytearray, a
        dictionary, or another Event
        """

        msg = super().parse(msg)
        event_type = msg['args'].pop('type')
        event_class = get_event_class_by_type(event_type)

        args = msg['args'] if 'args' in msg else {}
        args['id'] = msg['id'] if 'id' in msg else cls._generate_id()
        args['target'] = msg['target'] if 'target' in msg else Config.get('device_id')
        args['origin'] = msg['origin'] if 'origin' in msg else Config.get('device_id')
        args['timestamp'] = msg['_timestamp'] if '_timestamp' in msg else time.time()
        return event_class(**args)

    @staticmethod
    def _generate_id():
        """Generate a unique event ID"""
        return ''.join([f'{random.randint(0, 255):02x}' for _ in range(16)])

    @staticmethod
    def _is_relational_filter(filter: dict) -> bool:
        """
        Check if a condition is a relational filter.

        For a condition to be a relational filter, it must have at least one
        key starting with `$`.
        """
        if not isinstance(filter, dict):
            return False
        return any(key.startswith('$') for key in filter)

    @staticmethod
    def __relational_filter_matches(filter: dict, value: Any) -> bool:
        """
        Return True if the conditions in the filter match the given event
        arguments.
        """
        for op, filter_val in filter.items():
            comparator = _event_filter_operators.get(op)
            assert comparator, f'Invalid operator: {op}'

            # If this is a numeric or string filter, and one of the two values
            # is null, return False - it doesn't make sense to run numeric or
            # string comparison with null values.
            if (op in _numeric_filter_operators or op in _string_filter_operators) and (
                filter_val is None or value is None
            ):
                return False

            # If this is a numeric-only or string-only filter, then the
            # operands' types should be consistent with the operator.
            if op in _numeric_filter_operators:
                try:
                    value = float(value)
                    filter_val = float(filter_val)
                except (ValueError, TypeError) as e:
                    raise AssertionError(
                        f'Could not convert either "{value}" nor "{filter_val} to a number'
                    ) from e
            elif op in _string_filter_operators:
                assert isinstance(filter_val, str) and isinstance(value, str), (
                    f'Expected two strings, got "{filter_val}" '
                    f'({type(filter_val)}) and "{value}" ({type(value)})'
                )

            if not comparator(value, filter_val):
                return False

        return True

    @classmethod
    def _relational_filter_matches(cls, filter: dict, value: Any) -> bool:
        is_match = False
        try:
            is_match = cls.__relational_filter_matches(filter, value)
        except AssertionError as e:
            logger.error('Invalid filter: %s', e)

        if not is_match:
            return False

        return True

    # pylint: disable=too-many-branches,too-many-return-statements
    def _matches_condition(
        self,
        condition: dict,
        event_args: dict,
        result: "EventMatchResult",
        match_scores: list,
    ) -> bool:
        for attr, condition_value in condition.items():
            if attr not in event_args:
                return False

            event_value = event_args[attr]
            if isinstance(event_value, str):
                if self._is_relational_filter(condition_value):
                    if not self._relational_filter_matches(
                        condition_value, event_value
                    ):
                        return False
                else:
                    self._matches_argument(
                        argname=attr,
                        condition_value=condition_value,
                        event_args=event_args,
                        result=result,
                    )

                    if result.is_match:
                        match_scores.append(result.score)
                    else:
                        return False
            elif isinstance(condition_value, dict):
                if self._is_relational_filter(condition_value):
                    if not self._relational_filter_matches(
                        condition_value, event_value
                    ):
                        return False
                else:
                    if not isinstance(event_value, dict):
                        return False

                    if not self._matches_condition(
                        condition=condition_value,
                        event_args=event_value,
                        result=result,
                        match_scores=match_scores,
                    ):
                        return False
            else:
                if event_value != condition_value:
                    return False

                match_scores.append(2.0)

        return True

    def matches_condition(self, condition):
        """
        If the event matches an event condition, it will return an EventMatchResult
        :param condition: The platypush.event.hook.EventCondition object
        """

        result = EventMatchResult(is_match=False, parsed_args=self.args)
        match_scores = []

        if not isinstance(self, condition.type):
            return result

        if not self._matches_condition(
            condition=condition.args,
            event_args=self.args,
            result=result,
            match_scores=match_scores,
        ):
            return result

        result.is_match = True
        if match_scores:
            result.score = sum(match_scores) / float(len(match_scores))

        return result

    def _matches_argument(
        self, argname, condition_value, event_args, result: "EventMatchResult"
    ):
        """
        Returns an EventMatchResult if the event argument [argname] matches
        [condition_value].
        """

        # Simple equality match by default. It can be overridden by the derived classes.
        result.is_match = event_args.get(argname) == condition_value
        if result.is_match:
            result.score += 2
        else:
            result.score = 0

        return result

    def as_dict(self):
        """
        Converts the event into a dictionary
        """
        args = copy.deepcopy(self.args)
        flatten(args)
        return {
            'type': 'event',
            'target': self.target,
            'origin': self.origin if hasattr(self, 'origin') else None,
            'id': self.id if hasattr(self, 'id') else None,
            '_timestamp': self.timestamp,
            'args': {'type': self.type, **args},
        }

    def __str__(self):
        """
        Overrides the str() operator and converts
        the message into a UTF-8 JSON string
        """
        args = copy.deepcopy(self.args)
        flatten(args)
        return json.dumps(self.as_dict(), cls=self.Encoder)


@dataclass
class EventMatchResult:
    """
    When comparing an event against an event condition, you want to
    return this object. It contains the match status (True or False),
    any parsed arguments, and a match_score that identifies how "strong"
    the match is - in case of multiple event matches, the ones with the
    highest score will win.
    """

    is_match: bool
    score: float = 0.0
    parsed_args: dict = field(default_factory=dict)


def flatten(args):
    """
    Flatten a nested dictionary for string serialization.
    """
    if isinstance(args, dict):
        for key, value in args.items():
            if isinstance(value, date):
                args[key] = value.isoformat()
            elif isinstance(value, (dict, list)):
                flatten(args[key])
    elif isinstance(args, list):
        for i, arg in enumerate(args):
            if isinstance(arg, date):
                args[i] = arg.isoformat()
            elif isinstance(arg, (dict, list)):
                flatten(args[i])


_event_filter_operators = {
    '$gt': lambda a, b: a > b,
    '$gte': lambda a, b: a >= b,
    '$lt': lambda a, b: a < b,
    '$lte': lambda a, b: a <= b,
    '$eq': lambda a, b: a == b,
    '$ne': lambda a, b: a != b,
    '$regex': lambda a, b: re.search(b, a),
}

_numeric_filter_operators = {'$gt', '$gte', '$lt', '$lte'}

_string_filter_operators = {'$regex'}


# vim:sw=4:ts=4:et:



import multiprocessing
import queue
import socket
import threading
from typing import Optional

from platypush.backend import Backend
from platypush.message import Message


class TcpBackend(Backend):
    """
    Backend that reads messages from a configured TCP port.

    You can use this backend to send messages to Platypush from any TCP client, for example:

      .. code-block:: bash

        $ echo '{"type": "request", "action": "shell.exec", "args": {"cmd": "ls /"}}' | nc localhost 1234

    .. warning:: Be **VERY** careful when exposing this backend to the Internet. Unlike the HTTP backend, this backend
        doesn't implement any authentication mechanisms, so anyone who can connect to the TCP port will be able to
        execute commands on your Platypush instance.

    """

    # Maximum length of a request to be processed
    _MAX_REQ_SIZE = 2048

    def __init__(self, port, bind_address=None, listen_queue=5, **kwargs):
        """
        :param port: TCP port number
        :type port: int

        :param bind_address: Specify a bind address if you want to hook the
            service to a specific interface (default: listen for any connections).
        :type bind_address: str

        :param listen_queue: Maximum number of queued connections (default: 5)
        :type listen_queue: int
        """

        super().__init__(name=self.__class__.__name__, **kwargs)

        self.port = port
        self.bind_address = bind_address or '0.0.0.0'
        self.listen_queue = listen_queue
        self._accept_queue = multiprocessing.Queue()
        self._srv: Optional[multiprocessing.Process] = None

    def _process_client(self, sock, address):
        def _f():
            processed_bytes = 0
            open_brackets = 0
            msg = b''
            prev_ch = None

            while not self.should_stop():
                if processed_bytes > self._MAX_REQ_SIZE:
                    self.logger.warning(
                        'Ignoring message longer than {} bytes from {}'.format(
                            self._MAX_REQ_SIZE, address[0]
                        )
                    )
                    return

                ch = sock.recv(1)
                processed_bytes += 1

                if ch == b'':
                    break

                if ch == b'{' and prev_ch != b'\\':
                    open_brackets += 1

                if not open_brackets:
                    continue

                msg += ch

                if ch == b'}' and prev_ch != b'\\':
                    open_brackets -= 1

                if not open_brackets:
                    break

                prev_ch = ch

            if msg == b'':
                return

            msg = Message.build(msg)
            self.logger.info('Received request from %s: %s', msg, address[0])
            self.on_message(msg)

            response = self.get_message_response(msg)
            self.logger.info('Processing response on the TCP backend: %s', response)

            if response:
                sock.send(str(response).encode())

        def _f_wrapper():
            try:
                _f()
            finally:
                sock.close()

        threading.Thread(target=_f_wrapper, name='TCPListener').start()

    def _accept_process(self, serv_sock: socket.socket):
        while not self.should_stop():
            try:
                (sock, address) = serv_sock.accept()
                self._accept_queue.put((sock, address))
            except socket.timeout:
                continue

    def run(self):
        super().run()
        self.register_service(port=self.port)

        serv_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serv_sock.bind((self.bind_address, self.port))
        serv_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        serv_sock.settimeout(0.5)

        self.logger.info(
            'Initialized TCP backend on port %s with bind address %s',
            self.port,
            self.bind_address,
        )

        serv_sock.listen(self.listen_queue)
        self._srv = multiprocessing.Process(
            target=self._accept_process, args=(serv_sock,)
        )
        self._srv.start()

        while not self.should_stop():
            try:
                sock, address = self._accept_queue.get(timeout=1)
            except (socket.timeout, queue.Empty):
                continue

            self.logger.info('Accepted connection from client %s', address[0])
            self._process_client(sock, address)

        if self._srv:
            self._srv.kill()
            self._srv.join()
            self._srv = None

        self.logger.info('TCP backend terminated')


# vim:sw=4:ts=4:et:



import datetime
from typing import List, Dict, Any, Optional, Union, Tuple

import requests

from platypush.context import Variable
from platypush.message.event.foursquare import FoursquareCheckinEvent
from platypush.plugins import RunnablePlugin, action


class FoursquarePlugin(RunnablePlugin):
    """
    Plugin to interact with the `Foursquare Places API <https://developer.foursquare.com/docs/api>`_.

    It also raises events when a new check-in occurs on the user's account.

    In order to enable the Foursquare API on your account you need to:

        - Create a new app on the `Foursquare developers website <https://foursquare.com/developers/signup>`_.
        - Copy the ``client_id`` and ``client_secret``.
        - Add a redirect URL. It must point to a valid IP/hostname with a web server running, even if it runs
            locally. You can also use the local URL of the platypush web server - e.g. http://192.168.1.2:8008/.
        - Open the following URL:
            ``https://foursquare.com/oauth2/authenticate?client_id=CLIENT_ID&response_type=token&redirect_uri=REDIRECT_URI``.
            Replace ``CLIENT_ID`` and ``REDIRECT_URI`` with the parameters from your app.
        - Allow the application. You will be redirected to the URL you provided. Copy the ``access_token`` provided in
            the URL.

    """

    api_base_url = 'https://api.foursquare.com/v2'
    _last_created_at_varname = '_foursquare_checkin_last_created_at'
    _http_timeout = 10

    def __init__(self, access_token: str, poll_interval: float = 120, **kwargs):
        """
        :param access_token: The access token to use to authenticate to the Foursquare API.
        """
        super().__init__(poll_interval=poll_interval, **kwargs)
        self.access_token = access_token
        self._last_created_at = Variable(self._last_created_at_varname)

    def _get_url(self, endpoint):
        return (
            self.api_base_url
            + '/'
            + endpoint
            + '?oauth_token='
            + self.access_token
            + '&v='
            + datetime.date.today().strftime('%Y%m%d')
        )

    def _get_checkins(self) -> List[Dict[str, Any]]:
        url = self._get_url('users/self/checkins')
        return (
            requests.get(url, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('checkins', {})
            .get('items', [])
        )

    @action
    def get_checkins(self) -> List[Dict[str, Any]]:
        """
        Get the list of check-ins of the current user.
        :return: A list of checkins, as returned by the Foursquare API.
        """
        return self._get_checkins()

    @action
    def search(
        self,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        altitude: Optional[float] = None,
        latlng_accuracy: Optional[float] = None,
        altitude_accuracy: Optional[float] = None,
        near: Optional[str] = None,
        query: Optional[str] = None,
        limit: Optional[int] = None,
        url: Optional[str] = None,
        categories: Optional[List[str]] = None,
        radius: Optional[int] = None,
        sw: Optional[Union[Tuple[float], List[float]]] = None,
        ne: Optional[Union[Tuple[float], List[float]]] = None,
    ) -> List[Dict[str, Any]]:
        """
        Search for venues.

        :param latitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param longitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param near: Search near this place (e.g. "Chicago, IL" or "Amsterdam, NL"). Note either
            ``latitude, longitude`` or ``near`` should be provided.
        :param altitude: Search near this altitude in meters.
        :param latlng_accuracy:  Latitude/longitude accuracy in meters.
        :param altitude_accuracy: Altitude accuracy in meters.
        :param query: Search query (e.g. "coffee shops" or "restaurants").
        :param limit: Maximum number of results.
        :param url: Venue URL to search.
        :param categories: List of `category IDs <https://developer.foursquare.com/docs/resources/categories>`_
            to be searched.
        :param radius: Search radius in meters.
        :param sw: South/west boundary box as a ``[latitude, longitude]`` pair.
        :param ne: North/east boundary box as a ``[latitude, longitude]`` pair.
        :return: A list of venues, as returned by the Foursquare API.
        """
        assert (
            latitude and longitude
        ) or near, 'Specify either latitude/longitude or near'
        args = {}

        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if near:
            args['near'] = near
        if altitude:
            args['alt'] = altitude
        if latlng_accuracy:
            args['llAcc'] = latlng_accuracy
        if altitude_accuracy:
            args['altAcc'] = altitude_accuracy
        if query:
            args['query'] = query
        if limit:
            args['limit'] = int(limit)
        if url:
            args['url'] = url
        if categories:
            args['categoryId'] = ','.join(categories)
        if radius:
            args['radius'] = radius
        if sw:
            args['sw'] = sw
        if ne:
            args['ne'] = ne

        return (
            requests.get(
                self._get_url('venues/search'), params=args, timeout=self._http_timeout
            )
            .json()
            .get('response', {})
            .get('venues', [])
        )

    @action
    def explore(
        self,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        altitude: Optional[float] = None,
        latlng_accuracy: Optional[float] = None,
        altitude_accuracy: Optional[float] = None,
        section: Optional[str] = None,
        near: Optional[str] = None,
        query: Optional[str] = None,
        limit: Optional[int] = None,
        categories: Optional[List[str]] = None,
        radius: Optional[int] = None,
        open_now: bool = True,
        sort_by_distance: Optional[bool] = None,
        sort_by_popularity: Optional[bool] = None,
        price: Optional[List[int]] = None,
        saved: Optional[bool] = None,
    ) -> List[Dict[str, Any]]:
        """
        Explore venues around a location.

        :param latitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param longitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param near: Search near this place (e.g. "Chicago, IL" or "Amsterdam, NL"). Note either
            ``latitude, longitude`` or ``near`` should be provided.
        :param altitude: Search near this altitude in meters.
        :param latlng_accuracy:  Latitude/longitude accuracy in meters.
        :param altitude_accuracy: Altitude accuracy in meters.
        :param section: Section to search. Supported values:

            - food
            - drinks
            - coffee
            - shops
            - arts
            - outdoors
            - sights
            - trending
            - nextVenues

        :param query: Search query (e.g. "coffee shops" or "restaurants"). The parameter has no effect if
            ``section`` is specified.
        :param limit: Maximum number of results.
        :param categories: List of `category IDs <https://developer.foursquare.com/docs/resources/categories>`_
            to be searched.
        :param radius: Search radius in meters.
        :param open_now: Filter by open/not open now.
        :param sort_by_distance: Sort by distance.
        :param sort_by_popularity: Sort by popularity
        :param price: Price ranges, within the range ``[1,2,3,4]``.
        :param saved: Filter by saved/unsaved venues.

        :return: A list of venues, as returned by the Foursquare API.
        """
        assert (
            latitude and longitude
        ) or near, 'Specify either latitude/longitude or near'
        args = {}

        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if near:
            args['near'] = near
        if altitude:
            args['alt'] = altitude
        if latlng_accuracy:
            args['llAcc'] = latlng_accuracy
        if altitude_accuracy:
            args['altAcc'] = altitude_accuracy
        if section:
            args['section'] = section
        if query:
            args['query'] = query
        if categories:
            args['categoryId'] = ','.join(categories)
        if limit:
            args['limit'] = int(limit)
        if radius:
            args['radius'] = radius
        if open_now is not None:
            args['openNow'] = int(open_now)
        if sort_by_distance is not None:
            args['sortByDistance'] = int(sort_by_distance)
        if sort_by_popularity is not None:
            args['sortByPopularity'] = sort_by_popularity
        if saved is not None:
            args['saved'] = int(saved)
        if price:
            args['price'] = ','.join([str(p) for p in price])

        url = self._get_url('venues/explore')
        return (
            requests.get(url, params=args, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('venues', [])
        )

    @action
    def trending(
        self,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        near: Optional[str] = None,
        limit: Optional[int] = None,
        radius: Optional[int] = None,
    ) -> List[Dict[str, Any]]:
        """
        Get the trending venues around a location.

        :param latitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param longitude: Search near this latitude. Note either ``latitude, longitude`` or ``near`` should be provided.
        :param near: Search near this place (e.g. "Chicago, IL" or "Amsterdam, NL"). Note either
            ``latitude, longitude`` or ``near`` should be provided.
        :param limit: Maximum number of results.
        :param radius: Search radius in meters.

        :return: A list of venues, as returned by the Foursquare API.
        """
        assert (
            latitude and longitude
        ) or near, 'Specify either latitude/longitude or near'
        args = {}

        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if near:
            args['near'] = near
        if limit:
            args['limit'] = int(limit)
        if radius:
            args['radius'] = radius

        url = self._get_url('venues/trending')
        return (
            requests.get(url, params=args, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('venues', [])
        )

    @staticmethod
    def _parse_time(t):
        if isinstance(t, (int, float)):
            return datetime.datetime.fromtimestamp(t)
        if isinstance(t, str):
            return datetime.datetime.fromisoformat(t)

        assert isinstance(
            t, datetime.datetime
        ), f'Cannot parse object of type {type(t)} into datetime: {t}'
        return t

    @action
    def time_series(
        self,
        venue_id: Union[str, List[str]],
        start_at: Union[int, float, datetime.datetime, str],
        end_at: Union[int, float, datetime.datetime, str],
    ) -> List[Dict[str, Any]]:
        """
        Get the visitors stats about one or multiple venues over a time range. The user must be a manager of
        those venues.

        :param venue_id: Venue ID or list of IDs to get the stats for.
        :param start_at: Stats start time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.
        :param end_at: Stats end time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.

        :return: A list of venues, as returned by the Foursquare API.
        """
        if isinstance(venue_id, list):
            venue_id = ','.join(venue_id)

        args = {
            'venueId': venue_id,
            'startAt': self._parse_time(start_at),
            'endAt': self._parse_time(end_at),
        }

        url = self._get_url('venues/timeseries')
        return (
            requests.get(url, params=args, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('venues', [])
        )

    @action
    def stats(
        self,
        venue_id: str,
        start_at: Union[int, float, datetime.datetime, str],
        end_at: Union[int, float, datetime.datetime, str],
    ) -> List[Dict[str, Any]]:
        """
        Get the stats about a venue over a time range. The user must be a manager of that venue.

        :param venue_id: Venue ID.
        :param start_at: Stats start time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.
        :param end_at: Stats end time. Can be a UNIX timestamp, a datetime object or an ISO format datetime.

        :return: A list of venues, as returned by the Foursquare API.
        """
        args = {
            'startAt': self._parse_time(start_at).isoformat(),
            'endAt': self._parse_time(end_at).isoformat(),
        }

        url = self._get_url(f'venues/{venue_id}/stats')
        return (
            requests.get(url, params=args, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('venues', [])
        )

    @action
    def managed(self) -> List[Dict[str, Any]]:
        """
        Get the list of venues managed by the user.
        :return: A list of venues, as returned by the Foursquare API.
        """
        url = self._get_url('venues/managed')
        return (
            requests.get(url, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('venues', [])
            .get('items', [])
        )

    @action
    def checkin(
        self,
        venue_id: str,
        latitude: Optional[float] = None,
        longitude: Optional[float] = None,
        altitude: Optional[float] = None,
        latlng_accuracy: Optional[float] = None,
        altitude_accuracy: Optional[float] = None,
        shout: Optional[str] = None,
        broadcast: Optional[List[str]] = None,
    ) -> Dict[str, Any]:
        """
        Create a new check-in.

        :param venue_id: ID of the venue to check-in.
        :param latitude: Check-in latitude.
        :param longitude: Check-in longitude.
        :param altitude: Check-in altitude.
        :param latlng_accuracy:  Latitude/longitude accuracy in meters.
        :param altitude_accuracy: Altitude accuracy in meters.
        :param shout: Add a custom message to the check-in.
        :param broadcast: List of Visibility/share types of the check-in. Default: ``public``. Possible values are:

            - ``private``
            - ``public``
            - ``followers``
            - ``facebook``
            - ``twitter``

        :return: Foursquare API response.
        """
        args = {'venueId': venue_id}
        if latitude and longitude:
            args['ll'] = ','.join([str(latitude), str(longitude)])
        if altitude:
            args['alt'] = str(altitude)
        if latlng_accuracy:
            args['llAcc'] = str(latlng_accuracy)
        if altitude_accuracy:
            args['altAcc'] = str(altitude_accuracy)
        if shout:
            args['shout'] = shout
        if broadcast:
            args['broadcast'] = (
                ','.join(broadcast) if isinstance(broadcast, list) else broadcast
            )

        url = self._get_url('checkins/add')
        return (
            requests.post(url, data=args, timeout=self._http_timeout)
            .json()
            .get('response', {})
            .get('checkin', {})
        )

    def main(self):
        while not self.should_stop():
            checkins = self._get_checkins()
            if not checkins:
                return

            last_checkin = checkins[0]
            last_checkin_created_at = last_checkin.get('createdAt', 0)
            last_created_at = float(self._last_created_at.get() or 0)
            if last_created_at and last_checkin_created_at <= last_created_at:
                return

            self._bus.post(FoursquareCheckinEvent(checkin=last_checkin))
            self._last_created_at.set(last_checkin_created_at)
            self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from datetime import datetime
from typing import Iterable, Optional

from platypush.message.event import Event


class NewFeedEntryEvent(Event):
    """
    Event triggered when a new (RSS/Atom) feed entry is received.
    """

    def __init__(
        self,
        *,
        feed_url: str,
        url: str,
        title: Optional[str] = None,
        id: Optional[str] = None,
        feed_title: Optional[str] = None,
        published: Optional[datetime] = None,
        summary: Optional[str] = None,
        content: Optional[str] = None,
        author: Optional[str] = None,
        tags: Optional[Iterable[str]] = None,
        **kwargs
    ):
        super().__init__(
            feed_url=feed_url,
            url=url,
            title=title,
            id=id,
            feed_title=feed_title,
            published=published,
            summary=summary,
            content=content,
            author=author,
            tags=tags or [],
            **kwargs
        )



from collections import defaultdict
import hashlib
import io
import json
import threading
from typing import Any, Dict, Iterable, Optional, IO

import paho.mqtt.client as mqtt

from platypush.config import Config
from platypush.context import get_bus
from platypush.message import Message
from platypush.message.event.mqtt import MQTTMessageEvent
from platypush.message.request import Request
from platypush.plugins import RunnablePlugin, action
from platypush.utils import get_message_response

from ._client import DEFAULT_TIMEOUT, MqttCallback, MqttClient


class MqttPlugin(RunnablePlugin):
    """
    This plugin allows you to send custom message to a message queue compatible
    with the MQTT protocol, see https://mqtt.org/
    """

    def __init__(
        self,
        host: Optional[str] = None,
        port: int = 1883,
        topics: Optional[Iterable[str]] = None,
        tls_cafile: Optional[str] = None,
        tls_certfile: Optional[str] = None,
        tls_keyfile: Optional[str] = None,
        tls_version: Optional[str] = None,
        tls_ciphers: Optional[str] = None,
        tls_insecure: bool = False,
        username: Optional[str] = None,
        password: Optional[str] = None,
        client_id: Optional[str] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        run_topic_prefix: Optional[str] = None,
        listeners: Optional[Iterable[dict]] = None,
        **kwargs,
    ):
        """
        :param host: If set, MQTT messages will by default routed to this host
            unless overridden in `send_message` (default: None)
        :param port: If a default host is set, specify the listen port
            (default: 1883)
        :param topics: If a default ``host`` is specified, then this list will
            include a default list of topics that should be subscribed on that
            broker at startup.
        :param tls_cafile: If a default host is set and requires TLS/SSL,
            specify the certificate authority file (default: None)
        :param tls_certfile: If a default host is set and requires TLS/SSL,
            specify the certificate file (default: None)
        :param tls_keyfile: If a default host is set and requires TLS/SSL,
            specify the key file (default: None)
        :param tls_version: If TLS/SSL is enabled on the MQTT server and it
            requires a certain TLS version, specify it here (default: None).
            Supported versions: ``tls`` (automatic), ``tlsv1``, ``tlsv1.1``,
            ``tlsv1.2``.
        :param tls_ciphers: If a default host is set and requires TLS/SSL,
            specify the supported ciphers (default: None)
        :param tls_insecure: Set to True to ignore TLS insecure warnings
            (default: False).
        :param username: If a default host is set and requires user
            authentication, specify the username ciphers (default: None)
        :param password: If a default host is set and requires user
            authentication, specify the password ciphers (default: None)
        :param client_id: ID used to identify the client on the MQTT server
            (default: None). If None is specified then
            ``Config.get('device_id')`` will be used.
        :param timeout: Client timeout in seconds (default: 30 seconds).
        :param run_topic_prefix: If specified, the MQTT plugin will listen for
            messages on a topic in the format `{run_topic_prefix}/{device_id}.
            When a message is received, it will interpret it as a JSON request
            to execute, in the format
            ``{"type": "request", "action": "plugin.action", "args": {...}}``.

            .. warning:: This parameter is mostly kept for backwards
                compatibility, but you should avoid it - unless the MQTT broker
                is on a personal safe network that you own, or it requires
                user authentication and it uses SSL. The reason is that the
                messages received on this topic won't be subject to token
                verification, allowing unauthenticated arbitrary command
                execution on the target host. If you still want the ability of
                running commands remotely over an MQTT broker, then you may
                consider creating a dedicated topic listener with an attached
                event hook on
                :class:`platypush.message.event.mqtt.MQTTMessageEvent`. The
                hook can implement whichever authentication logic you like.

        :param listeners: If specified, the MQTT plugin will listen for
            messages on these topics. Use this parameter if you also want to
            listen on other MQTT brokers other than the primary one. This
            parameter supports a list of maps, where each item supports the
            same arguments passed to the main configuration (host, port, topic,
            password etc.). If host/port are omitted, then the host/port value
            from the plugin configuration will be used. If any of the other
            fields are omitted, then their default value will be used (usually
            null). Example:

                .. code-block:: yaml

                    listeners:
                        # This listener use the default configured host/port
                        - topics:
                              - topic1
                              - topic2
                              - topic3

                        # This will use a custom MQTT broker host
                        - host: sensors
                          port: 11883
                          username: myuser
                          password: secret
                          topics:
                              - topic4
                              - topic5

        """
        super().__init__(**kwargs)

        self.client_id = client_id or str(Config.get('device_id'))
        self.run_topic = (
            f'{run_topic_prefix}/{Config.get("device_id")}'
            if type(self) == MqttPlugin and run_topic_prefix
            else None
        )

        self._listeners_lock = defaultdict(threading.RLock)
        self.listeners: Dict[str, MqttClient] = {}  # client_id -> MqttClient map
        self.timeout = timeout
        self.default_listener = (
            self._get_client(
                host=host,
                port=port,
                topics=(
                    (tuple(topics) if topics else ())
                    + ((self.run_topic,) if self.run_topic else ())
                ),
                on_message=self.on_mqtt_message(),
                tls_cafile=tls_cafile,
                tls_certfile=tls_certfile,
                tls_keyfile=tls_keyfile,
                tls_version=tls_version,
                tls_ciphers=tls_ciphers,
                tls_insecure=tls_insecure,
                username=username,
                password=password,
                client_id=client_id,
                timeout=timeout,
            )
            if host
            else None
        )

        for listener in listeners or []:
            self._get_client(
                **self._mqtt_args(on_message=self.on_mqtt_message(), **listener)
            )

    def _get_client_id(
        self,
        host: str,
        port: int,
        client_id: Optional[str] = None,
        topics: Iterable[str] = (),
        **_,
    ) -> str:
        """
        Calculates a unique client ID given an MQTT configuration.
        """
        client_id = client_id or self.client_id
        client_hash = hashlib.sha1(
            '|'.join(
                [
                    self.__class__.__name__,
                    host,
                    str(port),
                    json.dumps(sorted(topics)),
                ]
            ).encode()
        ).hexdigest()

        return f'{client_id}-{client_hash}'

    def _mqtt_args(
        self,
        host: Optional[str] = None,
        port: int = 1883,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        topics: Iterable[str] = (),
        **kwargs,
    ):
        """
        :return: An MQTT configuration mapping that uses either the specified
            arguments (if host is specified), or falls back to the default
            configurated arguments.
        """
        default_conf = (
            self.default_listener.configuration if self.default_listener else {}
        )

        if not host:
            assert (
                self.default_listener
            ), 'No host specified and no configured default host'

            return {
                **default_conf,
                'topics': (*self.default_listener.topics, *topics),
            }

        return {
            'host': host,
            'port': port,
            'timeout': timeout or default_conf.get('timeout'),
            'topics': topics,
            **kwargs,
        }

    def on_mqtt_message(self) -> MqttCallback:
        """
        Default MQTT message handler. It forwards a
        :class:`platypush.message.event.mqtt.MQTTMessageEvent` event to the
        bus.
        """

        def handler(client: MqttClient, _, msg: mqtt.MQTTMessage):
            data = msg.payload
            try:
                data = data.decode('utf-8')
                data = json.loads(data)
            except (TypeError, AttributeError, ValueError):
                # Not a serialized JSON
                pass

            if self.default_listener and msg.topic == self.run_topic:
                try:
                    app_msg = Message.build(data)
                    self.on_exec_message(client, app_msg)
                except Exception as e:
                    self.logger.warning(
                        'Message execution error: %s: %s', type(e).__name__, str(e)
                    )
            else:
                get_bus().post(
                    MQTTMessageEvent(
                        host=client.host, port=client.port, topic=msg.topic, msg=data
                    )
                )

        return handler

    def on_exec_message(self, client: MqttClient, msg):
        """
        Message handler for (legacy) application requests over MQTT.
        """

        def response_thread(req: Request):
            """
            A separate thread to handle the response to a request.
            """
            if not self.run_topic:
                return

            response = get_message_response(req)
            if not response:
                return

            response_topic = f'{self.run_topic}/responses/{req.id}'
            self.logger.info(
                'Processing response on the MQTT topic %s: %s',
                response_topic,
                response,
            )

            client.publish(payload=str(response), topic=response_topic)

        self.logger.info('Received message on the MQTT backend: %s', msg)

        try:
            get_bus().post(msg)
        except Exception as e:
            self.logger.exception(e)
            return

        if isinstance(msg, Request):
            threading.Thread(
                target=response_thread,
                name='MQTTProcessorResponseThread',
                args=(msg,),
            ).start()

    def _get_client(
        self,
        host: Optional[str] = None,
        port: int = 1883,
        topics: Iterable[str] = (),
        client_id: Optional[str] = None,
        on_message: Optional[MqttCallback] = None,
        **kwargs,
    ) -> MqttClient:
        """
        :return: A :class:`platypush.message.event.mqtt.MqttClient` instance.
          It will return the existing client with the given inferred ID if it
          already exists, or it will register a new one.
        """
        if host:
            kwargs['host'] = host
            kwargs['port'] = port
        else:
            assert (
                self.default_listener
            ), 'No host specified and no configured default host'
            kwargs = self.default_listener.configuration

        on_message = on_message or self.on_mqtt_message()
        kwargs.update(
            {
                'topics': topics,
                'on_message': on_message,
                'client_id': client_id,
            }
        )

        client_id = self._get_client_id(
            host=kwargs['host'],
            port=kwargs['port'],
            client_id=client_id,
            topics=topics,
        )

        kwargs['client_id'] = client_id
        with self._listeners_lock[client_id]:
            client = self.listeners.get(client_id)
            if not (client and client.is_alive()):
                client = self.listeners[
                    client_id
                ] = MqttClient(  # pylint: disable=E1125
                    **kwargs
                )

        if topics:
            client.subscribe(*topics)

        return client

    @action
    def publish(
        self,
        topic: str,
        msg: Any,
        qos: int = 0,
        reply_topic: Optional[str] = None,
        **mqtt_kwargs,
    ):
        """
        Sends a message to a topic.

        :param topic: Topic/channel where the message will be delivered
        :param msg: Message to be sent. It can be a list, a dict, or a Message
            object.
        :param qos: Quality of Service (_QoS_) for the message - see `MQTT QoS
            <https://assetwolf.com/learn/mqtt-qos-understanding-quality-of-service>`_
            (default: 0).
        :param reply_topic: If a ``reply_topic`` is specified, then the action
            will wait for a response on this topic.
        :param mqtt_kwargs: MQTT broker configuration (host, port, username,
            password etc.). See :meth:`.__init__` parameters.
        """
        response_buffer = io.BytesIO()
        client = None

        try:
            # Try to parse it as a Platypush message or dump it to JSON from a dict/list
            if isinstance(msg, (dict, list)):
                msg = json.dumps(msg)

                try:
                    msg = Message.build(json.loads(msg))
                except (KeyError, TypeError, ValueError):
                    pass

            client = self._get_client(**mqtt_kwargs)
            client.connect()
            response_received = threading.Event()

            # If it's a request, then wait for the response
            if (
                isinstance(msg, Request)
                and self.default_listener
                and client.host == self.default_listener.host
                and self.run_topic
                and topic == self.run_topic
            ):
                reply_topic = f'{self.run_topic}/responses/{msg.id}'

            if reply_topic:
                client.on_message = self._response_callback(
                    reply_topic=reply_topic,
                    event=response_received,
                    buffer=response_buffer,
                )
                client.subscribe(reply_topic)

            client.publish(topic, str(msg), qos=qos)
            if not reply_topic:
                return None

            client.loop_start()
            ok = response_received.wait(timeout=client.timeout)
            if not ok:
                raise TimeoutError('Response timed out')

            return response_buffer.getvalue()
        finally:
            response_buffer.close()

            if client:
                client.stop()
                del client

    @action
    def subscribe(self, topic: str, **mqtt_kwargs):
        """
        Programmatically subscribe to a topic on an MQTT broker.

        Messages received on this topic will trigger a
        :class:`platypush.message.event.mqtt.MQTTMessageEvent` event that you
        can subscribe to.

        :param topic: Topic to subscribe to.
        :param mqtt_kwargs: MQTT broker configuration (host, port, username,
            password etc.). See :meth:`.__init__` parameters.
        """
        client = self._get_client(
            topics=(topic,), on_message=self.on_mqtt_message(), **mqtt_kwargs
        )

        if not client.is_alive():
            client.start()

    @action
    def unsubscribe(self, topic: str, **mqtt_kwargs):
        """
        Programmatically unsubscribe from a topic on an MQTT broker.

        :param topic: Topic to unsubscribe from.
        :param mqtt_kwargs: MQTT broker configuration (host, port, username,
            password etc.). See :meth:`.__init__` parameters.
        """
        client_id = self._get_client_id(
            topics=(topic,),
            **mqtt_kwargs,
        )

        with self._listeners_lock[client_id]:
            client = self.listeners.get(client_id)

        if not client:
            self.logger.info('No subscriptions found for topic %s', topic)
            return

        client.unsubscribe(topic)
        client.stop()
        del client

    def _response_callback(
        self, reply_topic: str, event: threading.Event, buffer: IO[bytes]
    ):
        """
        A response callback that writes the response to an IOBuffer and stops
        the client loop.
        """

        def on_message(client, _, msg):
            if msg.topic != reply_topic:
                return

            try:
                buffer.write(msg.payload)
                client.loop_stop()
            except Exception as e:
                self.logger.warning(
                    'Could not write the response back to the MQTT client: %s', e
                )
            finally:
                event.set()

        return on_message

    @action
    def send_message(self, *args, **kwargs):
        """
        Legacy alias for :meth:`platypush.plugins.mqtt.MqttPlugin.publish`.
        """
        return self.publish(*args, **kwargs)

    def main(self):
        if self.run_topic:
            self.logger.warning(
                'The MQTT integration is listening for commands on the topic %s.\n'
                'This approach is unsafe, as it allows any client to run unauthenticated requests.\n'
                'Please only enable it in test/trusted environments.',
                self.run_topic,
            )

        for listener in self.listeners.values():
            listener.start()

        self.wait_stop()

    def stop(self):
        """
        Disconnect all the clients upon plugin stop.
        """
        for listener in self.listeners.values():
            listener.stop()

        super().stop()

        for listener in self.listeners.values():
            try:
                listener.join(timeout=1)
                del listener
            except Exception:
                pass


# vim:sw=4:ts=4:et:



import enum
import logging
import re
from functools import wraps

from queue import LifoQueue
from typing import Optional

from ..common import exec_wrapper
from ..config import Config
from ..message.request import Request
from ..message.response import Response

logger = logging.getLogger('platypush')


class Statement(enum.Enum):
    """
    Enumerates the possible statements in a procedure.
    """

    BREAK = 'break'
    CONTINUE = 'continue'
    RETURN = 'return'


class Procedure:
    """Procedure class. A procedure is a pre-configured list of requests"""

    def __init__(self, name, _async, requests, args=None, backend=None):
        """
        Params:
            name     -- Procedure name
            _async   -- Whether the actions in the procedure are supposed to
                        be executed sequentially or in parallel (True or False)
            requests -- List of platypush.message.request.Request objects
        """

        self.name = name
        self._async = _async
        self.requests = requests
        self.backend = backend
        self.args = args or {}
        self._should_return = False

        for req in requests:
            req.backend = self.backend

    @classmethod
    # pylint: disable=too-many-branches,too-many-statements
    def build(
        cls,
        name,
        _async,
        requests,
        args=None,
        backend=None,
        id=None,  # pylint: disable=redefined-builtin
        procedure_class=None,
        **kwargs,
    ):
        reqs = []
        for_count = 0
        while_count = 0
        if_count = 0
        if_config = LifoQueue()
        procedure_class = procedure_class or cls
        key = None

        for request_config in requests:
            # Check if it's a break/continue/return statement
            if isinstance(request_config, str):
                reqs.append(Statement(request_config))
                continue

            # Check if this request is an if-else
            if len(request_config.keys()) >= 1:
                key = list(request_config.keys())[0]
                m = re.match(r'\s*(if)\s+\${(.*)}\s*', key)

                if m:
                    if_count += 1
                    if_name = f'{name}__if_{if_count}'
                    condition = m.group(2)

                    if_config.put(
                        {
                            'name': if_name,
                            '_async': False,
                            'requests': request_config[key],
                            'condition': condition,
                            'else_branch': [],
                            'backend': backend,
                            'id': id,
                        }
                    )

                    continue

                if key == 'else':
                    if if_config.empty():
                        raise RuntimeError(
                            f'else statement with no associated if in {name}'
                        )

                    conf = if_config.get()
                    conf['else_branch'] = request_config[key]
                    if_config.put(conf)

            if not if_config.empty():
                reqs.append(IfProcedure.build(**(if_config.get())))
                if key == 'else':
                    continue

            # Check if this request is a for loop
            if len(request_config.keys()) == 1:
                key = list(request_config.keys())[0]
                m = re.match(r'\s*(fork?)\s+([\w\d_]+)\s+in\s+(.*)\s*', key)

                if m:
                    for_count += 1
                    loop_name = f'{name}__for_{for_count}'

                    # A 'for' loop is synchronous. Declare a 'fork' loop if you
                    # want to process the elements in the iterable in parallel
                    _async = m.group(1) == 'fork'
                    iterator_name = m.group(2)
                    iterable = m.group(3)

                    loop = ForProcedure.build(
                        name=loop_name,
                        _async=_async,
                        requests=request_config[key],
                        backend=backend,
                        id=id,
                        iterator_name=iterator_name,
                        iterable=iterable,
                    )

                    reqs.append(loop)
                    continue

            # Check if this request is a while loop
            if len(request_config.keys()) == 1:
                key = list(request_config.keys())[0]
                m = re.match(r'\s*while\s+\${(.*)}\s*', key)

                if m:
                    while_count += 1
                    loop_name = f'{name}__while_{while_count}'
                    condition = m.group(1).strip()

                    loop = WhileProcedure.build(
                        name=loop_name,
                        _async=False,
                        requests=request_config[key],
                        condition=condition,
                        backend=backend,
                        id=id,
                    )

                    reqs.append(loop)
                    continue

            request_config['origin'] = Config.get('device_id')
            request_config['id'] = id
            if 'target' not in request_config:
                request_config['target'] = request_config['origin']

            request = Request.build(request_config)
            reqs.append(request)

        while not if_config.empty():
            pending_if = if_config.get()
            reqs.append(IfProcedure.build(**pending_if))

        return procedure_class(
            name=name,
            _async=_async,
            requests=reqs,
            args=args,
            backend=backend,
            **kwargs,
        )

    @staticmethod
    def _find_nearest_loop(stack):
        for proc in stack[::-1]:
            if isinstance(proc, LoopProcedure):
                return proc

        raise AssertionError('break/continue statement found outside of a loop')

    # pylint: disable=too-many-branches,too-many-statements
    def execute(self, n_tries=1, __stack__=None, **context):
        """
        Execute the requests in the procedure.

        :param n_tries: Number of tries in case of failure before raising a RuntimeError.
        """
        if not __stack__:
            __stack__ = [self]
        else:
            __stack__.append(self)

        if self.args:
            args = self.args.copy()
            for k, v in args.items():
                v = Request.expand_value_from_context(v, **context)
                args[k] = v
                context[k] = v
            logger.info('Executing procedure %s with arguments %s', self.name, args)
        else:
            logger.info('Executing procedure %s', self.name)

        response = Response()
        token = Config.get('token')

        for request in self.requests:
            if callable(request):
                response = request(**context)
                continue

            if isinstance(request, Statement):
                if request == Statement.RETURN:
                    self._should_return = True
                    for proc in __stack__:
                        proc._should_return = True  # pylint: disable=protected-access
                    break

                if request in [Statement.BREAK, Statement.CONTINUE]:
                    loop = self._find_nearest_loop(__stack__)
                    if request == Statement.BREAK:
                        loop._should_break = True  # pylint: disable=protected-access
                    else:
                        loop._should_continue = True  # pylint: disable=protected-access
                    break

            should_continue = getattr(self, '_should_continue', False)
            should_break = getattr(self, '_should_break', False)
            if isinstance(self, LoopProcedure) and (should_continue or should_break):
                if should_continue:
                    self._should_continue = (  # pylint: disable=attribute-defined-outside-init
                        False
                    )

                break

            if token and not isinstance(request, Statement):
                request.token = token

            context['_async'] = self._async
            context['n_tries'] = n_tries
            exec_ = getattr(request, 'execute', None)
            if callable(exec_):
                response = exec_(__stack__=__stack__, **context)

            if not self._async and response:
                if isinstance(response.output, dict):
                    for k, v in response.output.items():
                        context[k] = v

                context['output'] = response.output
                context['errors'] = response.errors

            if self._should_return:
                break

        return response or Response()

    def to_dict(self):
        return {
            'name': self.name,
            'requests': self.requests,
            'args': self.args,
            '_async': self._async,
        }


class LoopProcedure(Procedure):
    """
    Base class while and for/fork loops.
    """

    def __init__(self, name, requests, _async=False, args=None, backend=None):
        super().__init__(
            name=name, _async=_async, requests=requests, args=args, backend=backend
        )
        self._should_break = False
        self._should_continue = False


class ForProcedure(LoopProcedure):
    """
    Models a loop procedure, which expresses a construct similar to a
    for loop in a programming language. The 'for' keyword implies a synchronous
    loop, i.e. the nested actions will be executed in sequence. Use 'fork'
    instead of 'for' if you want to run the actions in parallel.

    Example::

        procedure.sync.process_results:
            - action: http.get
              args:
                  url: https://some-service/some/json/endpoint
                  # Example response: { "results": [ {"id":1, "name":"foo"}, {"id":2,"name":"bar"} ]}
            - for result in ${results}:
                - action: some.custom.action
                  args:
                      id: ${result['id']}
                      name: ${result['name']}

    """

    def __init__(
        self,
        name,
        iterator_name,
        iterable,
        requests,
        _async=False,
        args=None,
        backend=None,
    ):
        super().__init__(
            name=name, _async=_async, requests=requests, args=args, backend=backend
        )
        self.iterator_name = iterator_name
        self.iterable = iterable

    # pylint: disable=eval-used
    def execute(self, *_, **context):
        try:
            iterable = eval(self.iterable)
            assert hasattr(
                iterable, '__iter__'
            ), f'Object of type {type(iterable)} is not iterable: {iterable}'
        except Exception as e:
            logger.debug('Iterable %s expansion error: %s', self.iterable, e)
            iterable = Request.expand_value_from_context(self.iterable, **context)

        response = Response()

        for item in iterable:
            if self._should_return:
                logger.info('Returning from %s', self.name)
                break

            if self._should_continue:
                self._should_continue = False
                logger.info('Continuing loop %s', self.name)
                continue

            if self._should_break:
                self._should_break = False
                logger.info('Breaking loop %s', self.name)
                break

            context[self.iterator_name] = item
            response = super().execute(**context)

        return response


class WhileProcedure(LoopProcedure):
    """
    Models a while loop procedure.

    Example::

        procedure.process_results:
            - action: http.get
              args:
                  url: https://some-service/some/json/endpoint
                  # Example response: {"id":1, "name":"foo"}}

            - while ${output}:
                - action: some.custom.action
                  args:
                      id: ${id}
                      name: ${name}

                - action: http.get
                  args:
                      url: https://some-service/some/json/endpoint
                      # Example response: {"id":1, "name":"foo"}}

    """

    def __init__(
        self, name, condition, requests, _async=False, args=None, backend=None
    ):
        super().__init__(
            name=name, _async=_async, requests=requests, args=args, backend=backend
        )
        self.condition = condition

    @staticmethod
    def _get_context(**context):
        for k, v in context.items():
            try:
                context[k] = eval(v)  # pylint: disable=eval-used
            except Exception as e:
                logger.debug('Evaluation error for %s=%s: %s', k, v, e)
                if isinstance(v, str):
                    try:
                        context[k] = eval(  # pylint: disable=eval-used
                            '"' + re.sub(r'(^|[^\\])"', '\1\\"', v) + '"'
                        )
                    except Exception as ee:
                        logger.warning(
                            'Could not parse value for context variable %s=%s: %s',
                            k,
                            v,
                            ee,
                        )
                        logger.warning('Context: %s', context)
                        logger.exception(e)

        return context

    def execute(self, *_, **context):
        response = Response()
        context = self._get_context(**context)
        for k, v in context.items():
            locals()[k] = v

        while True:
            condition_true = eval(self.condition)  # pylint: disable=eval-used
            if not condition_true:
                break

            if self._should_return:
                logger.info('Returning from %s', self.name)
                break

            if self._should_continue:
                self._should_continue = False
                logger.info('Continuing loop %s', self.name)
                continue

            if self._should_break:
                self._should_break = False
                logger.info('Breaking loop %s', self.name)
                break

            response = super().execute(**context)

            if response.output and isinstance(response.output, dict):
                new_context = self._get_context(**response.output)
                for k, v in new_context.items():
                    locals()[k] = v

        return response


class IfProcedure(Procedure):
    """
    Models an if-else construct.

    Example::

        procedure.sync.process_results:
            - action: http.get
              args:
                url: https://some-service/some/json/endpoint
                # Example response: { "sensors": [ {"temperature": 18 } ] }

            - if ${sensors['temperature'] < 20}:
              - action: shell.exec
                args:
                  cmd: '/path/turn_on_heating.sh'
            - else:
              - action: shell.exec
                args:
                  cmd: '/path/turn_off_heating.sh'

    """

    def __init__(
        self,
        name,
        condition,
        requests,
        else_branch=None,
        args=None,
        backend=None,
        id=None,  # pylint: disable=redefined-builtin
        **kwargs,
    ):
        kwargs['_async'] = False
        self.condition = condition
        self.else_branch = else_branch
        reqs = []

        for req in requests:
            if isinstance(req, dict):
                req['origin'] = Config.get('device_id')
                req['id'] = id
                if 'target' not in req:
                    req['target'] = req['origin']

                req = Request.build(req)

            reqs.append(req)

        super().__init__(name=name, requests=reqs, args=args, backend=backend, **kwargs)

    @classmethod
    # pylint: disable=arguments-differ
    def build(
        cls,
        name,
        *_,
        condition,
        requests,
        args=None,
        backend=None,
        id=None,  # pylint: disable=redefined-builtin
        else_branch=None,
        **kwargs,
    ):
        kwargs['_async'] = False
        if else_branch:
            else_branch = super().build(
                name=name + '__else',
                requests=else_branch,
                args=args,
                backend=backend,
                id=id,
                procedure_class=Procedure,
                **kwargs,
            )

        return super().build(
            name=name,
            condition=condition,
            requests=requests,
            else_branch=else_branch,
            args=args,
            backend=backend,
            id=id,
            **kwargs,
        )

    def execute(self, *_, **context):
        for k, v in context.items():
            locals()[k] = v

        condition_true = eval(self.condition)  # pylint: disable=eval-used
        response = Response()

        if condition_true:
            response = super().execute(**context)
        elif self.else_branch:
            response = self.else_branch.execute(**context)

        return response


def procedure(name_or_func: Optional[str] = None, *upper_args, **upper_kwargs):
    name = name_or_func if isinstance(name_or_func, str) else None

    def func_wrapper(f):
        """
        Public decorator to mark a function as a procedure.
        """

        f.procedure = True
        f.procedure_name = name

        @wraps(f)
        def _execute_procedure(*args, **kwargs):
            args = [*upper_args, *args]
            kwargs = {**upper_kwargs, **kwargs}
            return exec_wrapper(f, *args, **kwargs)

        return _execute_procedure

    if callable(name_or_func):
        return func_wrapper(name_or_func)

    return func_wrapper


# vim:sw=4:ts=4:et:



from typing import Optional, Dict, Any

from platypush.plugins import action
from platypush.plugins.weather import WeatherPlugin
from platypush.schemas.weather.buienradar import (
    WeatherSchema,
    PrecipitationForecastSchema,
)


class WeatherBuienradarPlugin(WeatherPlugin):  # pylint: disable=too-many-ancestors
    """
    Plugin for getting weather updates through Buienradar - a Dutch weather
    app.
    """

    def __init__(self, lat: float, long: float, time_frame: int = 120, **kwargs):
        """
        :param lat: Default latitude.
        :param long: Default longitude.
        :param time_frame: Default number of minutes to look ahead for
            precipitation forecast.
        """
        super().__init__(**kwargs)
        self.lat = lat
        self.long = long
        self.time_frame = time_frame
        self.latest_bulletin = {}

    def get_data(
        self,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        time_frame: Optional[int] = None,
    ) -> Dict[str, Any]:
        from buienradar.buienradar import get_data, parse_data  # type: ignore
        from buienradar.constants import SUCCESS, CONTENT, RAINCONTENT, DATA  # type: ignore

        lat = lat or self.lat
        long = long or self.long
        time_frame = time_frame or self.time_frame

        result = get_data(latitude=lat, longitude=long)
        if not result.get(SUCCESS):
            raise RuntimeError('Error while retrieving data')

        data = result.get(CONTENT)
        rain_data = result.get(RAINCONTENT)
        result = parse_data(data, rain_data, lat, long, time_frame)
        return result.get(DATA, {})

    def _get_current_weather(
        self, *_, lat: Optional[float] = None, long: Optional[float] = None, **__
    ):
        """
        Get the current weather conditions.

        :param lat: Weather latitude (default: configured latitude)
        :param long: Weather longitude (default: configured longitude)
        :return: .. schema:: schemas.weather.buienradar.WeatherSchema
        """
        return WeatherSchema().dump(self.get_data(lat, long, 60))

    def _get_forecast(
        self, *_, lat: Optional[float] = None, long: Optional[float] = None, **__
    ):
        return [
            {
                'datetime': weather['datetime'].isoformat(),
                **dict(WeatherSchema().dump(weather)),
            }
            for weather in self.get_data(lat, long, 60).get('forecast', [])
        ]

    @action
    def get_precipitation(
        self,
        lat: Optional[float] = None,
        long: Optional[float] = None,
        time_frame: Optional[int] = None,
    ):
        """
        Get the precipitation forecast for the specified time frame.

        :param lat: Weather latitude (default: configured latitude)
        :param long: Weather longitude (default: configured longitude)
        :param time_frame: Time frame for the forecast in minutes (default: configured time_frame)
        """
        return PrecipitationForecastSchema().dump(
            self.get_data(lat, long, time_frame).get('precipitation_forecast', {})
        )


# vim:sw=4:ts=4:et:



from typing import Callable, Optional, Sequence, Union

import pvrhino

from platypush.message.event.assistant import (
    ConversationTimeoutEvent,
    IntentRecognizedEvent,
)

from ._base import BaseProcessor


class IntentProcessor(BaseProcessor):
    """
    Implementation of the speech-to-intent processor using the Picovoice Rhino
    engine.
    """

    def __init__(
        self, *args, get_rhino_args: Callable[[], dict] = lambda: {}, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._get_rhino_args = get_rhino_args
        # model_path -> Rhino instance cache
        self._rhino = {}

    @property
    def _model_path(self) -> Optional[str]:
        return self._get_rhino_args().get('model_path')

    @property
    def sample_rate(self) -> int:
        return self._get_rhino().sample_rate

    @property
    def frame_length(self) -> int:
        return self._get_rhino().frame_length

    def _get_rhino(self) -> pvrhino.Rhino:
        if not self._rhino.get(self._model_path):
            self._rhino[self._model_path] = pvrhino.create(**self._get_rhino_args())

        return self._rhino[self._model_path]

    def process(
        self, audio: Sequence[int]
    ) -> Optional[Union[IntentRecognizedEvent, ConversationTimeoutEvent]]:
        """
        Process the audio and return an ``IntentRecognizedEvent`` if the intent was
        understood, or a ``ConversationTimeoutEvent`` if the conversation timed
        out, or ``None`` if the intent processing is not yet finalized.
        """
        event = None
        rhino = self._get_rhino()
        self._ctx.is_final = rhino.process(audio)

        if self._ctx.is_final:
            inference = rhino.get_inference()
            self.logger.debug(
                'Intent detection finalized. Inference understood: %s',
                inference.is_understood,
            )

            if inference.is_understood:
                event = IntentRecognizedEvent(
                    intent=inference.intent,
                    slots=dict(inference.slots),
                )

        if not event and self._ctx.timed_out:
            event = ConversationTimeoutEvent()

        return event

    def reset(self) -> None:
        if not self._enabled:
            return

        with self._state_lock:
            self._get_rhino().reset()
            super().reset()

    def stop(self):
        if not self._enabled:
            return

        super().stop()

        with self._state_lock:
            objs = self._rhino.copy()
            for key, obj in objs.items():
                obj.delete()
                self._rhino.pop(key)



import enum
import math
import time
from typing import Dict, List

from platypush.common.sensors import Numeric
from platypush.entities.devices import Device
from platypush.entities.distance import DistanceSensor
from platypush.entities.sensors import NumericSensor, RawSensor
from platypush.plugins import action
from platypush.plugins.sensor import SensorPlugin


class Rotation(enum.IntEnum):
    """
    Enumerates the possible rotations of the sensor.
    """

    ROTATE_0 = 0
    ROTATE_90 = 90
    ROTATE_180 = 180
    ROTATE_270 = 270


class SPISlot(enum.Enum):
    """
    Enumeration for the possible SPI slot positions of the sensor on the
    Breakout Garden (front or back).
    """

    FRONT = 'front'
    BACK = 'back'


# pylint: disable=too-many-ancestors
class SensorPmw3901Plugin(SensorPlugin):
    """
    Plugin to interact with an `PMW3901 <https://github.com/pimoroni/pmw3901-python>`_
    optical flow and motion sensor.
    """

    def __init__(
        self,
        rotation=Rotation.ROTATE_0.value,
        spi_slot=SPISlot.FRONT.value,
        spi_port=0,
        poll_interval=0.01,
        **kwargs,
    ):
        """
        :param rotation: Rotation angle for the captured optical flow. Possible options: 0, 90, 180, 270 (default: 0)
        :type rotation: int

        :param spi_slot: SPI slot where the sensor is connected if you're using a Breakout Garden interface.
            Possible options: 'front', 'back' (default: 'front')
        :type spi_slot: str

        :param spi_port: SPI port (default: 0)
        :type spi_slot: int
        """
        from pmw3901 import BG_CS_FRONT_BCM, BG_CS_BACK_BCM

        super().__init__(poll_interval=poll_interval, **kwargs)

        self.spi_port = spi_port
        self._sensor = None
        self._events_per_sec = {}
        self.x, self.y = (0, 0)

        try:
            if isinstance(rotation, int):
                rotation = [r for r in Rotation if r.value == rotation][0]
            self.rotation = rotation
        except IndexError as e:
            raise AssertionError(
                f'{rotation} is not a valid value for rotation - '
                f'possible values: {[r.value for r in Rotation]}'
            ) from e

        try:
            if isinstance(spi_slot, str):
                spi_slot = [s for s in SPISlot if s.value == spi_slot][0]

            if spi_slot == SPISlot.FRONT:
                self.spi_slot = BG_CS_FRONT_BCM
            else:
                self.spi_slot = BG_CS_BACK_BCM
        except IndexError as e:
            raise AssertionError(
                f'{spi_slot} is not a valid value for spi_slot - '
                f'possible values: {[s.value for s in SPISlot]}'
            ) from e

    def _get_sensor(self):
        from pmw3901 import PMW3901

        if not self._sensor:
            self._sensor = PMW3901(
                spi_port=self.spi_port, spi_cs=1, spi_cs_gpio=self.spi_slot
            )
            self._sensor.set_rotation(self.rotation)

        return self._sensor

    @action
    def get_measurement(self, *_, **__):
        """
        :returns: dict. Example:

        .. code-block:: python

           output = {
               "motion_x": 3,   # Detected motion vector X-coord
               "motion_y": 4,   # Detected motion vector Y-coord
               "motion_mod": 5  # Detected motion vector module
               "motion_events_per_sec": 7  # Number of motion events detected in the last second
           }

        """

        sensor = self._get_sensor()

        while True:
            try:
                x, y = sensor.get_motion()
                break
            except RuntimeError:
                self.wait_stop(0.01)

        secs = int(time.time())
        if (x, y) != (self.x, self.y):
            (self.x, self.y) = (x, y)

            if secs not in self._events_per_sec:
                self._events_per_sec = {secs: 1}
            else:
                self._events_per_sec[secs] += 1

        return {
            'motion_x': x,
            'motion_y': y,
            'motion_mod': math.sqrt(x * x + y * y),
            'motion_events_per_sec': self._events_per_sec.get(secs, 0),
        }

    def transform_entities(self, entities: Dict[str, Numeric]) -> List[Device]:
        return [
            Device(
                id='pmw3901',
                name='PMW3901',
                children=[
                    RawSensor(
                        id='pmw3901:motion',
                        name='motion',
                        description='Motion vector',
                        value=[entities['motion_x'], entities['motion_y']],
                        is_json=True,
                    ),
                    DistanceSensor(
                        id='pmw3901:module',
                        name='module',
                        description='Motion vector module',
                        value=entities['motion_mod'],
                    ),
                    NumericSensor(
                        id='pmw3901:rate',
                        name='rate',
                        description='Events per second',
                        value=entities['motion_events_per_sec'],
                    ),
                ],
            )
        ]


# vim:sw=4:ts=4:et:



from time import time
from typing import Optional

from platypush.context import get_bus
from platypush.message.event.clipboard import ClipboardEvent
from platypush.plugins import RunnablePlugin, action


class ClipboardPlugin(RunnablePlugin):
    """
    Plugin to programmatically copy strings to your system clipboard,
    monitor and get the current clipboard content.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_text: Optional[str] = None

    @action
    def copy(self, text):
        """
        Copies a text to the OS clipboard

        :param text: Text to copy
        :type text: str
        """
        import pyclip

        pyclip.copy(text)

    @action
    def paste(self):
        """
        Get the current content of the clipboard
        """
        import pyclip

        return pyclip.paste(text=True)

    def main(self):
        import pyclip

        last_error_time = 0

        while not self.should_stop():
            try:
                text = pyclip.paste(text=True)
                if text and text != self._last_text:
                    get_bus().post(ClipboardEvent(text=text))
                    self._last_text = text
            except Exception as e:
                if time() - last_error_time > 60:
                    last_error_time = time()
                    self.logger.error(
                        'Could not access the clipboard: %s: %s', type(e), e
                    )
            finally:
                self._should_stop.wait(0.1)

        self.logger.info('Stopped clipboard monitor backend')


# vim:sw=4:ts=4:et:



from typing import Optional

import aioxmpp
import aioxmpp.im.p2p

from platypush.message.event.xmpp import (
    XmppConversationAddedEvent,
    XmppConversationEnterEvent,
    XmppConversationExitEvent,
    XmppConversationJoinEvent,
    XmppConversationLeaveEvent,
)

from ._base import XmppBaseHandler


# pylint: disable=too-many-ancestors
class XmppConversationHandler(XmppBaseHandler):
    """
    Handler for XMPP conversation events.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.conversation: aioxmpp.im.ConversationService = self._client.summon(
            aioxmpp.im.ConversationService
        )
        self.conversation.on_conversation_added.connect(self._on_conversation_added)  # type: ignore

    def _on_conversation_added(
        self, conversation: aioxmpp.im.p2p.Conversation, *_, **__
    ):
        if not isinstance(conversation, aioxmpp.im.p2p.Conversation):
            return  # Don't add signals to rooms - they'll have their own

        conversation_id = str(conversation.jid)
        if not self._state.conversations.get(conversation_id):
            self._register_conversation_events(conversation)
            self._state.users.add(conversation_id)
            self._state.conversations[conversation_id] = conversation
            self._post_conversation_event(
                XmppConversationAddedEvent,
                conversation=conversation,
                members=[
                    str(m.direct_jid if m.direct_jid else m.conversation_jid)
                    for m in conversation.members
                ],
            )

    def _register_conversation_events(self, conversation: aioxmpp.im.p2p.Conversation):
        if not isinstance(conversation, aioxmpp.im.p2p.Conversation):
            return  # Don't add signals to rooms - they'll have their own

        conversation.on_enter.connect(self._on_conversation_enter(conversation))  # type: ignore
        conversation.on_exit.connect(self._on_conversation_exit(conversation))  # type: ignore
        conversation.on_join.connect(self._on_conversation_join(conversation))  # type: ignore
        conversation.on_leave.connect(self._on_conversation_leave(conversation))  # type: ignore

    def _on_conversation_enter(self, conversation: aioxmpp.im.p2p.Conversation):
        def callback(*_, **__):
            self._post_conversation_event(XmppConversationEnterEvent, conversation)

        return callback

    def _on_conversation_exit(self, conversation: aioxmpp.im.p2p.Conversation):
        def callback(*_, **__):
            self._post_conversation_event(XmppConversationExitEvent, conversation)

        return callback

    def _on_conversation_join(self, conversation: aioxmpp.im.p2p.Conversation):
        def callback(member: aioxmpp.im.p2p.Member, *_, **__):
            self._post_conversation_member_event(
                XmppConversationJoinEvent, conversation=conversation, member=member
            )

        return callback

    def _on_conversation_leave(self, conversation: aioxmpp.im.p2p.Conversation):
        def callback(member: aioxmpp.im.p2p.Member, *_, **__):
            if member.is_self:
                user_id = str(conversation.jid)
                # Remove the conversation from the map of active conversations
                self._state.conversations.pop(user_id, None)
                self._state.users = self._state.users.difference({user_id})

            self._post_conversation_member_event(
                XmppConversationLeaveEvent,
                conversation=conversation,
                member=member,
            )

        return callback

    def send_message(
        self,
        user_id: str,
        body: str,
        language: Optional[str] = None,
    ):
        lang = language or self._lang
        msg = aioxmpp.Message(
            type_=aioxmpp.MessageType.CHAT, to=aioxmpp.JID.fromstr(user_id)
        )
        msg.body.update({lang: body})
        self._client.enqueue(msg)
